/**
  *  \file client/chartloc.cc
  */

#include "client/chartloc.h"
#include "game/univ.h"
#include "game/global.h"
#include "game/ship.h"

GChartLocation::GChartLocation()
    : location(0, 0),
      pid(0),
      list_index(0),
      changed(false)
{
    viewport.center.x = viewport.center.y = 0;
}

GChartLocation::~GChartLocation()
{ }

void
GChartLocation::updateLocation()
{
    // Build object lists
    GUniverse& univ = getDisplayedTurn().getCurrentUniverse();
    pid = univ.getPlanetAt(location);
    list.clear();
    list.addObjectsAt(univ, location, ls_IncludeForeignShips + ls_IncludePlanet, 0);
    list.postprocessUser(false);

    // Reset list index
    list_index = (list.getNumObjects() > 0 ? 1 : 0);
}

/** Initialize before first load. */
void
GChartLocation::init()
{
    viewport.center.x = viewport.center.y = 2000;
    location.x = location.y = 2000;
    list.clear();
    pid = 0;
    changed = true;
}

/** Initialize after game load. */
void
GChartLocation::initAfterLoad()
{
    init();
    updateLocation();
}

/** Initialize upon starchart entry. */
void
GChartLocation::initEntry(const GfxRect& area)
{
    /* initialize viewport */
    viewport.mult = viewport.divi = 1;
    viewport.view = area;
    viewport.vec_id = 0;
    viewport.drawing_filter = viewport.NO_FILTER;
    viewport.init();

    /* cause initial redraw */
    changed = true;
}

/** Set center of chart. */
void
GChartLocation::setCenter(GPoint pt)
{
    pt = pt.getNormalized();

    // Same as last time?
    if (pt == viewport.center)
        return;

    // Change center
    viewport.center = pt;
    changed = true;
}

/** Set location. */
void
GChartLocation::setLocation(GPoint pt)
{
    pt = pt.getNormalized();

    // Same as last time?
    if (pt == location)
        return;

    // Change location
    location = pt;
    changed = true;
    updateLocation();
}

/** Get currently-locked object. */
GObjectId
GChartLocation::getCurrentObject() const
{
    if (list_index > 0 && list_index <= list.getNumObjects())
        return list.getObjectId(list_index); /* 1-based access! */
    else
        return GObjectId();
}

/** Set currently-locked object. Also adjusts the position if required. */
void
GChartLocation::setCurrentObject(GObject& obj)
{
    // If it's a map object, go there
    if (GMapObject* mo = dynamic_cast<GMapObject*>(&obj)) {
        setLocation(mo->getPos());
    }

    // Focus it.
    int index = list.getIndexFor(obj);
    if (index > 0 && index != list_index) {
        list_index = index;
        changed = true;
    }
}

/** Get Id of currently-locked ship, if any.
    \return ship id, 0 if no ship locked */
int
GChartLocation::getCurrentShipId() const
{
    GObjectId id = getCurrentObject();
    if (!id.isGameObject())
        return 0;

    GShip* sh = dynamic_cast<GShip*>(&id.getObject());
    if (!sh)
        return 0;

    return sh->getId();
}

/** Perform zooming. */
void
GChartLocation::doZoom(bool enlarge)
{
    if (enlarge)
        changed |= viewport.zoomIn(getDisplayedTurn().getChartBBox());
    else
        changed |= viewport.zoomOut(getDisplayedTurn().getChartBBox());
}

/** Scroll through objects at current location. */
void
GChartLocation::doScroll(int direction, bool mark)
{
    int new_index;
    if (direction > 0)
        new_index = list.getNext(list_index, mark);
    else
        new_index = list.getPrev(list_index, mark);

    if (new_index != list_index) {
        list_index = new_index;
        changed = true;
    }
}

/** Toggle options. */
void
GChartLocation::toggleOptions(int32 which)
{
    getChartOpts(false, viewport.mult, viewport.divi).toggleOptions(which);
    changed = true;
}

bool
GChartLocation::isChanged() const
{
    return changed;
}

void
GChartLocation::markChanged()
{
    changed = true;
}

void
GChartLocation::clearChanged()
{
    changed = false;
}
