/**
  *  \file client/cm-distance.cc
  *  \brief Starchart mode: Distance
  */

#include <cmath>
#include "client/cm-distance.h"
#include "client/chartloc.h"
#include "cpluslib/format.h"
#include "game/formula.h"
#include "gfx/gfxutil.h"
#include "gfx/palette.h"
#include "ui/event.h"
#include "ui/uiglobal.h"
#include "util/nls.h"

/** Given distance and speed, compute estimated time. */
static int
ETA(double d, int warp)
{
    int32 r = roundToInt(d);
    int32 way = warp*warp;

    return (r + way-1) / way;
}

/** Constructor.
    \param origin Origin of measurement
    \param ship_id Predict fuel usage for this ship (0=none) */
WDistanceChartMode::WDistanceChartMode(GPoint origin, int ship_id)
    : WSimpleChartMode(ChartMode_PrimaryMode),
      origin(origin),
      ship_id(ship_id)
{ }

/** Destructor. */
WDistanceChartMode::~WDistanceChartMode()
{ }

bool
WDistanceChartMode::handleEvent(const UIEvent& event, GChartLocation& loc)
{
    if (event.type == ev_Key) {
        switch (event.key.code) {
         case 'x':
            // Swap ends
            {
                GPoint p = origin;
                origin = loc.getLocation();
                loc.setLocation(p);
            }
            return true;

         // case 'f':
         //    // Use ship for fuel consumption

         // case 'w':
         // case ss_Ctrl + 'w':
         //    // Set ship waypoint
         //    return true;

         case SDLK_BACKSPACE:
            // Back to beginning and exit
            loc.setLocation(origin);
            delete this;
            return true;

         case SDLK_DELETE:
         case 'd':
            // Exit mode
            delete this;
            return true;
        }
    }
    return WSimpleChartMode::handleEvent(event, loc);
}

void
WDistanceChartMode::drawOverlays(GfxCanvas& can, GChartLocation& loc)
{
    GfxRect pos = loc.getViewport().view;

    GfxParameters parm;
    parm.color = standard_colors[COLOR_WHITE];
    parm.font = fonts[FONT_NORMAL];

    // Distance
    double dist = std::sqrt(double(distanceSquaredRaw(origin, loc.getLocation())));
    outText(can, parm, pos.x, pos.y,
            format(_("Distance from first point: %.1f ly"), dist));

    // Speeds, angle
    string_t speed_angle = format(_("Warp/Time: 6/%d 7/%d 8/%d 9/%d"),
                                  ETA(dist, 6),
                                  ETA(dist, 7),
                                  ETA(dist, 8),
                                  ETA(dist, 9));
    if (dist > 0)
        speed_angle += format(", %d",
                              roundToInt(getHeadingDeg(loc.getLocation().x - origin.x,
                                                       loc.getLocation().y - origin.y)));

    outText(can, parm, pos.x, pos.y + font_heights[FONT_NORMAL], speed_angle);

    // FIXME: ship fuel usage display, waypoint change

    WSimpleChartMode::drawOverlays(can, loc);
}

void
WDistanceChartMode::drawCursor(GfxCanvas& can, GChartLocation& loc)
{
    GfxParameters parm;
    parm.color = standard_colors[COLOR_WHITE];

    GfxPoint a = loc.getViewport().scaleXY(origin);
    GfxPoint b = loc.getViewport().scaleXY(loc.getLocation());

    drawLine(can, parm, a.x, a.y, b.x, b.y);

    WSimpleChartMode::drawCursor(can, loc);
}
