/**
  *  \file client/dlg-build.cc
  *
  *  (c) 2001-2009 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dlg-build.h"
#include "client/winfo.h"
#include "game/planet.h"
#include "game/player.h"
#include "game/spec.h"
#include "game/struct.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/table.h"
#include "ui/widgets.h"
#include "util/misc.h"
#include "util/nls.h"
#include "cpluslib/format.h"
#include "ui/msgbox.h"
#include "resmgr/resid.h"
#include "ui/util.h"
#include "client/dlg-component.h"
#include "game/global.h"

enum {
    cm_AddComponent    = 4000,
    cm_RemoveComponent = 4001,
    cm_CountChanged    = 4002,
    cm_BuildComponent  = 4003
};

/*************************** WComponentListbox ***************************/

/** \class WComponentListbox
    \brief List box containing list of ship parts

    This class only provides a method drawComponent() to draw ship
    parts. Derived classes use this to achieve a uniform look across
    all component lists. */

/** Contructor.
    \param id       widget id
    \param lines    number of lines of widget
    \param width    width of widget
    \param planet   planet. May be null. Ignored if it doesn't have a
                    starbase. Tech levels (and tick marks) are taken
                    from here. */
WComponentListbox::WComponentListbox(int id, int lines, int width, GPlanet* planet)
    : UIListbox(id, lines, width, 0), tick_planet(planet)
{
    if (!tick_planet->hasBase())
        tick_planet = 0;
}

/** Draw a component entry.
    \param can     target canvas
    \param line    logical line number (see drawEntry())
    \param com     component to draw.
    \param tech_area the tech area used by this component.
    \param have    amount of such components in storage (used for tick marks) */
void
WComponentListbox::drawComponent(GfxCanvas& can, int line, const GComponent& com, TTechLevel tech_area, int have)
{
    UIColor::Color c = (!tick_planet || com.getTech() <= tick_planet->getBaseTech(tech_area)
                        ? UIColor::tc_Static
                        : com.getTech() > 6 && !allowRegistered(tick_planet->getOwner())
                        ? UIColor::tc_Faded
                        : UIColor::tc_Blue);
    const char* tick = have ? "\7" : " ";
    drawEntry(can, c, line, tick + com.getName());
}

/****************************** WHullListbox *****************************/

/** \class WHullListbox
    \brief Hull List box

    This widget displays a list of hulls. You can (actually, must)
    provide a list of hulls to display. */

/** Constructor. \see WComponentListbox::WComponentListbox() */
WHullListbox::WHullListbox(int id, int lines, int width, GPlanet* p)
    : WComponentListbox(id, lines, width, p)
{ }

WHullListbox::~WHullListbox()
{ }

void
WHullListbox::drawPart(GfxCanvas& can, int from, int to)
{
    for (int i = from; i < to; ++i) {
        const int slot = i + getTop();
        if (slot < int(content.size())) {
            const int hull = content[slot];
            int have = 0;
            if (tick_planet && tick_planet->hasBase()) {
                int n = getTruehullSlot(tick_planet->getOwner(), hull);
                if (n)
                    have = tick_planet->getBaseHullStoreSlot(n);
            }
            drawComponent(can, i, getHull(hull), tl_Hull, have);
        } else {
            drawEntry(can, UIColor::tc_Static, i, string_t());
        }
    }
}

/** Add a hull to the end of the list.
    \param n hull number (1..NUM_HULLS)
    \see swapContents() */
void
WHullListbox::addHullId(int n)
{
    content.push_back(n);
    setSize(content.size());
    drawWidget();
}

/** Set content of this list.
    \param new_content vector of hull numbers (1..NUM_HULLS). After
    this function completes, this widget will display the old contents
    of /new_content/, and /new_content/ will contain the old content. */
void
WHullListbox::swapContents(std::vector<int>& new_content)
{
    int now = getCurrentHull();
    content.swap(new_content);
    setSize(content.size());
    scrollToHull(now);
    adjustPosition();
    drawWidget();
}

/** Get currently-selected hull number. */
int
WHullListbox::getCurrentHull() const
{
    if (getCurrent() < int(content.size()))
        return content[getCurrent()];
    else
        return 0;
}

/** Scroll to specified hull. Does nothing if this hull is not on our
    list. */
void
WHullListbox::scrollToHull(int id)
{
    for (std::vector<int>::size_type i = 0; i < content.size(); ++i) {
        if (id == content[i]) {
            scrollTo(i, 1);
            return;
        }
    }
}

/************ WEngineListbox / WLauncherListbox / WBeamListbox ***********/

WEngineListbox::WEngineListbox(int id, int width, GPlanet* tick_planet)
    : WComponentListbox(id, NUM_ENGINES, width, tick_planet)
{
    setSize(NUM_ENGINES);
}

void
WEngineListbox::drawPart(GfxCanvas& can, int from, int to)
{
    for (int i = from; i < to; ++i) {
        int slot = getTop() + i + 1;
        if (slot <= NUM_ENGINES)
            drawComponent(can, i, getEngine(slot), tl_Engine,
                          tick_planet ? tick_planet->getBaseEngineStore(slot) : 0);
        else
            drawEntry(can, UIColor::tc_Static, i, string_t());
    }
}

WBeamListbox::WBeamListbox(int id, int width, GPlanet* tick_planet)
    : WComponentListbox(id, NUM_BEAMS, width, tick_planet)
{
    setSize(NUM_BEAMS);
}

void
WBeamListbox::drawPart(GfxCanvas& can, int from, int to)
{
    for (int i = from; i < to; ++i) {
        int slot = getTop() + i + 1;
        if (slot <= NUM_BEAMS)
            drawComponent(can, i, getBeam(slot), tl_Beam,
                          tick_planet ? tick_planet->getBaseBeamStore(slot) : 0);
        else
            drawEntry(can, UIColor::tc_Static, i, string_t());
    }
}

WLauncherListbox::WLauncherListbox(int id, int width, GPlanet* tick_planet)
    : WComponentListbox(id, NUM_TORPS, width, tick_planet)
{
    setSize(NUM_TORPS);
}

void
WLauncherListbox::drawPart(GfxCanvas& can, int from, int to)
{
    for (int i = from; i < to; ++i) {
        int slot = getTop() + i + 1;
        if (slot <= NUM_TORPS)
            drawComponent(can, i, getTorp(slot), tl_Torpedo,
                          tick_planet ? tick_planet->getBaseLauncherStore(slot) : 0);
        else
            drawEntry(can, UIColor::tc_Static, i, string_t());
    }
}

/***************************** WOrderDisplay *****************************/

/** \class WOrderDisplay
    \brief Display a Build Order

    This will display a build order on the ship build screen. */
WOrderDisplay::WOrderDisplay()
{
    o.hull = 0;
}

void
WOrderDisplay::drawContent(GfxCanvas& can)
{
    UITableWriter tw(can, getSkin(), getExtent(), FONT_NORMAL, UITableWriter::fl_Fill);

    if (o.hull) {
        GHull& h = getHull(o.hull);

        tw.newLColumn();
        tw.writeVariableCell(UIColor::tc_Contrast, _("Your order:"));
        tw.writeVariableCell(UIColor::tc_Static, "\7 " + h.getName());
        tw.writeVariableCell(UIColor::tc_Static,
                             /* this format string turns into
                                "Transwarp Drive" or "7 x Transwarp Drive" */
                             format(_("\7 %!d%!1{%0$d x %}%1$s"),
                                    h.getNumEngines(),
                                    getEngine(o.engine).getName()));
        if (o.beam_count)
            tw.writeVariableCell(UIColor::tc_Static,
                                 format(_("\7 %!d%!1{%0$d x %}%1$s"),
                                        o.beam_count,
                                        getBeam(o.beam_type).getName()));
        if (o.torp_count)
            tw.writeVariableCell(UIColor::tc_Static,
                                 format(_("\7 %!d%!1{%0$d x %}%1$s"),
                                        o.torp_count,
                                        getTorp(o.torp_type).getName()));
        if (h.getNumBays())
            tw.writeVariableCell(UIColor::tc_Static,
                                 format(_("\7 %d fighter bay%!1{s%}"), h.getNumBays()));
    }
    tw.finish();
    can.update(getExtent());
}

void
WOrderDisplay::getLayoutInfo(LayoutInfo& info)
{
    info.flags = 0;
    info.min_size = info.pref_size = font_sizes[FONT_NORMAL].scaledBy(20, 5);
}

void
WOrderDisplay::setBuildOrder(const TBuildOrder& o)
{
    this->o = o;
    drawWidget();
}

static UIBaseWidget&
addFrame(ObjectHolder& h, UIBaseWidget& w)
{
    UIFrameGroup& fg = h.add(new UIFrameGroup(UIHBoxLayout::instance0));
    fg.add(w);
    return fg;
}

/************************** WPartStorageDisplay **************************/

WPartStorageDisplay::WPartStorageDisplay()
    : UIWidget(), n(0)
{ }

void
WPartStorageDisplay::drawContent(GfxCanvas& can)
{
    UISkin& s = getSkin();
    s.startDraw();
    s.setSolid(true);
    s.setColor(UIColor::tc_Static);
    s.outTextF(can, FONT_NORMAL, getExtent().x, getExtent().y, getExtent().w,
               _("In storage:"));
    s.outTextF(can, FONT_NORMAL, getExtent().x, getExtent().y + font_heights[FONT_NORMAL], getExtent().w,
               format(_("%d pcs."), n));
    can.update(getExtent());
}

void
WPartStorageDisplay::getLayoutInfo(LayoutInfo& info)
{
    info.min_size = info.pref_size = font_sizes[FONT_NORMAL].scaledBy(7, 2);
    info.flags = 0;
}

void
WPartStorageDisplay::setCount(int n)
{
    if (n != this->n) {
        this->n = n;
        drawWidget(false);
    }
}


WPartStorageGroup::WPartStorageGroup()
    : UIGroup(UIVBoxLayout::instance5),
      image(105+2, 93+2 /* FIXME! Make this variable? */),
      psd(),
      button(cm_BuildComponent, _("Spc-Build"), ' ')
{
    add(image);
    add(psd);
    // add(spacer);   FIXME: do we need it?
    add(button);
    button.addCommand();
}

WPartStorageGroup::~WPartStorageGroup()
{ }

/***************************** WPartCountInfo ****************************/

WPartCountInfo::WPartCountInfo(bool torps)
    : torps(torps), want(0), max(0)
{ }

void
WPartCountInfo::drawContent(GfxCanvas& can)
{
    const char* fmt = torps ? _("Max. launchers on ship: %d") : _("Max. beams on ship: %d");
    int x = getExtent().x, y = getExtent().y, wi = getExtent().w;

    UISkin& s = getSkin();
    s.startDraw();
    s.setSolid(true);
    s.setColor(UIColor::tc_White);
    s.outTextF(can, FONT_NORMAL, x, y, wi, format(fmt, max));
    y += font_heights[FONT_NORMAL];
    if (max > 0) {
        s.outTextF(can, FONT_NORMAL, x, y, wi, format(_("Actual number: %d"), want));
    } else {
        s.outTextF(can, FONT_NORMAL, x, y, wi, string_t());
    }
    can.update(getExtent());
}

void
WPartCountInfo::getLayoutInfo(LayoutInfo& info)
{
    info.min_size = info.pref_size = font_sizes[FONT_NORMAL].scaledBy(15, 2);
    info.flags = li_FlexibleH;
}

void
WPartCountInfo::setPartsMax(int n)
{
    if (n != max) {
        max = n;
        want = max;
        drawWidget(false);
    }
}

void
WPartCountInfo::setPartsWanted(int n)
{
    if (n != want) {
        want = n;
        drawWidget(false);
    }
}

WPartCountGroup::WPartCountGroup(bool torps)
    : UIGroup(UIHBoxLayout::instance0),
      pci(torps),
      button_group(UIVBoxLayout::instance0),
      plus(cm_AddComponent, "+", '+'),
      minus(cm_RemoveComponent, "-", '-')
{
    add(pci);
    add(button_group);

    button_group.add(plus);
    button_group.add(minus);

    plus.addCommand();
    plus.setState(UIButton::bf_SmallSize, true);
    minus.addCommand();
    minus.setState(UIButton::bf_SmallSize, true);
}

WPartCountGroup::~WPartCountGroup()
{ }

void
WPartCountGroup::setPartsMax(int n)
{
    pci.setPartsMax(n);
    setStateOf(cm_AddComponent, st_Disabled, n == 0);
    setStateOf(cm_RemoveComponent, st_Disabled, n == 0);
}

bool
WPartCountGroup::handleCommand(const UICommand& cmd, UIBaseWidget& sender)
{
    if (cmd.code == cm_AddComponent) {
        int n = getPartsWanted() + (cmd.arg ? cmd.arg : 1);
        if (n > getPartsMax())
            n = getPartsMax();
        setPartsWanted(n);
        handleCommand(UICommand(cm_CountChanged), *this);
        return true;
    } else if (cmd.code == cm_RemoveComponent) {
        int n = getPartsWanted() - (cmd.arg ? cmd.arg : 1);
        if (n < 0)
            n = 0;
        setPartsWanted(n);
        handleCommand(UICommand(cm_CountChanged), *this);
        return true;
    } else {
        return UIGroup::handleCommand(cmd, sender);
    }
}

/*************************** WEngineInfoDiagram **************************/

WEngineInfoDiagram::WEngineInfoDiagram()
    : UIWidget(0), engine(1)
{ }

void
WEngineInfoDiagram::drawContent(GfxCanvas& can)
{
    int bottom = getExtent().getBottomY();
    int center = getExtent().getCenter().x;

    GfxParameters& parm = getSkin();
    getSkin().startDraw();
    getSkin().setColor(UIColor::tc_Static);
    parm.just_x = 1;
    parm.just_y = 0;

    string_t desc = _("Fuel used per\nlightyear");
    
    getSkin().outText(can, FONT_SMALL, center, bottom - 2*font_heights[FONT_SMALL],
                      strFirst(desc, "\n"));
    strRemove(desc, "\n");
    getSkin().outText(can, FONT_SMALL, center, bottom - 1*font_heights[FONT_SMALL],
                      desc);

    GfxRect diagram_pos(getExtent());
    diagram_pos.h -= 2*font_heights[FONT_SMALL] + 1;

    drawFrameDown(can, diagram_pos);
    diagram_pos.grow(-1, -1);
    drawSolidBar(can, diagram_pos, standard_colors[COLOR_BLACK]);
    diagram_pos.grow(-1, -1);

    parm.just_x = 0;
    parm.color = standard_colors[COLOR_WHITE];
    parm.bgcolor = GFX_TRANSPARENT;
    parm.font = fonts[FONT_SMALL];
    outText(can, parm, diagram_pos.x, diagram_pos.getBottomY() - font_heights[FONT_SMALL],
            _("Warp \032"));
    diagram_pos.h -= font_heights[FONT_SMALL]+1;

    /* Diagram */
    int prev_ff = 0;
    int prev_x = 0;
    for (int warp = 1; warp <= 10; ++warp) {
        int new_x = warp*diagram_pos.w / 10;
        int new_ff = warp == 10 ? 1000 : getEngine(engine).getFuelFactor(warp) / (warp*warp);
        if (new_ff > 999)
            new_ff = 999;

        int delta_x = new_x - prev_x;

        /* draw the bars */
        for (int dx = 1; dx < delta_x; ++dx) {
            int this_ff    = (dx*new_ff + (delta_x-dx)*prev_ff) / delta_x;
            int this_color = COLOR_STATUS+15 - (16*this_ff)/1000;
            int this_y     = this_ff*diagram_pos.h/1000;
            parm.color = standard_colors[this_color];
            drawVLine(can, parm,
                      diagram_pos.x + dx + prev_x,
                      diagram_pos.getBottomY() - this_y,
                      diagram_pos.getBottomY());
        }
        prev_x = new_x;
        prev_ff = new_ff;
    }

    can.update(getExtent());
}

void
WEngineInfoDiagram::getLayoutInfo(LayoutInfo& info)
{
    /* The diagram is 2 (top) + 100 (size) + 1 (gap) + text + 2 (bottom)
       + 1 (gap) + 2*text */
    info.pref_size = info.min_size =
        GfxPoint(102, 106 + 3*font_heights[FONT_SMALL]);
    info.flags = 0;
}

void
WEngineInfoDiagram::setEngine(int n)
{
    if (n != engine) {
        engine = n;
        drawWidget(false);
    }
}

/*************************** WEngineStorageInfo **************************/

WEngineStorageInfo::WEngineStorageInfo()
    : UIWidget(), have(0), need(0)
{ }

void
WEngineStorageInfo::setEnginesAvailable(int n)
{
    if (n != have) {
        have = n;
        drawWidget(false);
    }
}

void
WEngineStorageInfo::setEnginesNeeded(int n)
{
    if (n != need) {
        need = n;
        drawWidget(false);
    }
}

void
WEngineStorageInfo::drawContent(GfxCanvas& can)
{
    int x = getExtent().x, y = getExtent().y, wi = getExtent().w;

    UISkin& s = getSkin();
    s.setSolid();

    s.setColor(UIColor::tc_White);
    s.outTextF(can, FONT_TITLE, x, y, wi, format(_("Ship requires %d engine%!1{s%}."), need));
    y += font_heights[FONT_TITLE];

    s.outTextF(can, FONT_NORMAL, x, y, wi, format(_("In storage: %d engine%!1{s%}."), have));
    y += font_heights[FONT_NORMAL];

    can.update(getExtent());
}

void
WEngineStorageInfo::getLayoutInfo(LayoutInfo& info)
{
    info.min_size = info.pref_size =
        font_sizes[FONT_NORMAL].scaledBy(15, 1).movedBy(0, font_heights[FONT_TITLE]);
    info.flags = 0;
}

/**************************** WShipBuildScreen ***************************/

/** \class WShipBuildScreen
    \brief Ship Build Screen, general version

    This class provides a complete ship build screen which allows the
    user to select a ship type. It does not provide buttons to confirm
    a build transaction. A derived class must add these. */

WShipBuildScreen::WShipBuildScreen(string_t name, GPlanet& the_planet)
    : UIWindow(name, UIWindow::BLUE_WINDOW, COLOR_DARK),
      hull_list(3000, 10, 15 * font_ems[FONT_NORMAL], &the_planet),
      engine_list(3001, 12 * font_ems[FONT_NORMAL], &the_planet),
      beam_list(3002, 12 * font_ems[FONT_NORMAL], &the_planet),
      torp_list(3003, 12 * font_ems[FONT_NORMAL], &the_planet),
      beam_count(false),
      torp_count(true),
      transfer(the_planet),
      action(the_planet, transfer),
      item_cost_dpy(action, false),
      cards(),
      tabs(cards),
      conn_update(the_planet.sig_changed.add(this, &WShipBuildScreen::onPlanetChanged))
{
    setLayout(UIVBoxLayout::instance5);

    add(h.add(new UIFocusIterator(1000, 1003, UIFocusIterator::fi_Horiz)));

    tabs.addTab(1000, UIKeyString(_("Starship Hulls")));
    tabs.addTab(1001, UIKeyString(_("Engines")));
    tabs.addTab(1002, UIKeyString(_("Beams")));
    tabs.addTab(1003, UIKeyString(_("Torpedoes")));

    /* Hulls */
    UIGroup& hull_group = h.add(new UIGroup(UIHBoxLayout::instance5, 1000));
    for (int i = 1; i <= NUM_HULLS_PER_PLAYER; ++i)
        if (int j = getTruehull(the_planet.getOwner(), i))
            hull_list.addHullId(j);
    hull_group.add(addFrame(h, hull_list));
    hull_group.add(hull_storage);
    hull_group.add(hull_info);
    hull_list.scrollToHull(action.getBuildOrder().hull);
    hull_list.sig_moved.add(this, &WShipBuildScreen::updateHullPage);

    /* Engines */
    UIGroup& engine_group = h.add(new UIGroup(UIHBoxLayout::instance5, 1001));
    UIGroup& engine_vgroup = h.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& engine_hgroup = h.add(new UIGroup(UIHBoxLayout::instance5));
    engine_group.add(addFrame(h, engine_list));
    engine_group.add(engine_diag);
    engine_vgroup.add(engine_info);
    engine_vgroup.add(h.add(new UISpacer()));
    engine_vgroup.add(engine_storage);
    engine_vgroup.add(engine_hgroup);
    engine_hgroup.add(h.add(new UIButton(cm_BuildComponent, _("Spc-Build"), ' ')).addCommand());
    engine_hgroup.add(h.add(new UISpacer()));
    engine_group.add(engine_vgroup);
    engine_list.scrollTo(action.getBuildOrder().engine - 1, 1);
    engine_list.sig_moved.add(this, &WShipBuildScreen::updateEnginePage);

    /* Beams */
    UIGroup& beam_group = h.add(new UIGroup(UIHBoxLayout::instance5, 1002));
    UIGroup& beam_vgroup = h.add(new UIGroup(UIVBoxLayout::instance0));
    beam_group.add(addFrame(h, beam_list));
    beam_group.add(beam_storage);
    beam_vgroup.add(beam_info);
    beam_vgroup.add(h.add(new UISpacer()));
    beam_vgroup.add(beam_count);
    beam_group.add(beam_vgroup);
    beam_list.scrollTo(action.getBuildOrder().beam_type - 1, 1);
    beam_list.sig_moved.add(this, &WShipBuildScreen::updateBeamPage);

    /* Torps */
    UIGroup& torp_group = h.add(new UIGroup(UIHBoxLayout::instance5, 1003));
    UIGroup& torp_vgroup = h.add(new UIGroup(UIVBoxLayout::instance0));
    torp_group.add(addFrame(h, torp_list));
    torp_group.add(torp_storage);
    torp_vgroup.add(torp_info);
    torp_vgroup.add(h.add(new UISpacer()));
    torp_vgroup.add(torp_count);
    torp_group.add(torp_vgroup);
    torp_list.scrollTo(action.getBuildOrder().torp_type - 1, 1);
    torp_list.sig_moved.add(this, &WShipBuildScreen::updateLauncherPage);

    cards.add(hull_group);
    cards.add(engine_group);
    cards.add(beam_group);
    cards.add(torp_group);
    cards.sig_focus_changed.add(this, &WShipBuildScreen::updateBuildOrder);

    add(tabs);
    add(cards);

    /* Bill */
    UIGroup& bill_group = h.add(new UIGroup(UIHBoxLayout::instance5));
    bill_group.add(item_cost_dpy);
    bill_group.add(order_dpy);
    add(bill_group);

    hull_list.setFocusedWidget();
    updateHullPage();
    updateEnginePage();
    updateBeamPage();
    updateLauncherPage();
}

WShipBuildScreen::~WShipBuildScreen()
{
    if (!action.isFinished())
        action.doCancel();
    if (!transfer.isFinished())
        transfer.doCancel();
}

bool
WShipBuildScreen::handleCommand(const UICommand& cmd, UIBaseWidget& sender)
{
    if (cmd.code == cm_CountChanged) {
        updateBuildOrder();
        return true;
    } else if (cmd.code == cm_BuildComponent) {
        GPlanet& p = action.getPlanet();
        switch (tabs.getCurrentTab()) {
         case 1000:
            if (int slot = getTruehullSlot(p.getOwner(), hull_list.getCurrentHull()))
                doBuildComponent(getDisplayedTurn(), p.getId(),
                                 tl_Hull, slot,
                                 getHull(hull_list.getCurrentHull()));
            break;
         case 1001:
            doBuildComponent(getDisplayedTurn(), p.getId(),
                             tl_Engine, engine_list.getCurrent()+1,
                             getEngine(engine_list.getCurrent()+1));
            break;
         case 1002:
            doBuildComponent(getDisplayedTurn(), p.getId(),
                             tl_Beam, beam_list.getCurrent()+1,
                             getBeam(beam_list.getCurrent()+1));
            break;
         case 1003:
            doBuildComponent(getDisplayedTurn(), p.getId(),
                             tl_Torpedo, torp_list.getCurrent()+1,
                             getTorp(torp_list.getCurrent()+1));
            break;
        }
        return true;
    } else {
        return UIWindow::handleCommand(cmd, sender);
    }
}

void
WShipBuildScreen::updateHullPage()
{
    const int hull = hull_list.getCurrentHull();
    const GPlanet& p = action.getPlanet();

    hull_info.setHullId(hull);
    if (int n = getTruehullSlot(p.getOwner(), hull))
        hull_storage.setCount(p.getBaseHullStoreSlot(n));
    else
        hull_storage.setCount(0);
    hull_storage.setImage(requestPixmap(ResId(res::ship, getHull(hull).getPicture(), hull)));

    engine_storage.setEnginesNeeded(getHull(hull).getNumEngines());
    beam_count.setPartsMax(getHull(hull).getMaxBeams());
    torp_count.setPartsMax(getHull(hull).getMaxLaunchers());

    updateBuildOrder();
}

void
WShipBuildScreen::updateBeamPage()
{
    const int beam = beam_list.getCurrent()+1;
    beam_info.setBeamId(beam);
    beam_storage.setCount(action.getPlanet().getBaseBeamStore(beam));
    beam_storage.setImage(requestPixmap(ResId("beam", beam)));
    updateBuildOrder();
}

void
WShipBuildScreen::updateLauncherPage()
{
    const int torp = torp_list.getCurrent()+1;
    torp_info.setTorpId(torp);
    torp_storage.setCount(action.getPlanet().getBaseLauncherStore(torp));
    torp_storage.setImage(requestPixmap(ResId("launcher", torp)));
    updateBuildOrder();
}

void
WShipBuildScreen::updateEnginePage()
{
    const int engine = engine_list.getCurrent()+1;
    const GPlanet& p = action.getPlanet();
    engine_info.setEngineId(engine);
    engine_diag.setEngine(engine);
    engine_storage.setEnginesAvailable(p.getBaseEngineStore(engine));
    updateBuildOrder();
}

void
WShipBuildScreen::updateBuildOrder()
{
    TBuildOrder bo;

    if (action.isFinished())  // !
        return;

    bo.hull   = hull_list.getCurrentHull();
    bo.engine = engine_list.getCurrent()+1;
    bo.beam_type  = beam_list.getCurrent()+1;
    bo.beam_count = beam_count.getPartsWanted();
    bo.torp_type  = torp_list.getCurrent()+1;
    bo.torp_count = torp_count.getPartsWanted();
    bo.unused = 0;
    action.setBuildOrder(bo);
    order_dpy.setBuildOrder(bo);

    // since this might be called during destruction, we cannot use
    // 'tabs.getCurrentTab()'. 'tabs' might already be gone.
    if (UIBaseWidget* p = cards.getFocusedChild()) {
        switch (p->getId()) {
         case 1000: {
            item_cost_dpy.setCost(getHull(hull_list.getCurrentHull()).getCost());
            item_cost_dpy.setTech(action.getPlanet().getBaseTech(tl_Hull),
                                  getHull(hull_list.getCurrentHull()).getTech());
            break;
         }
         case 1001: {
            GEngine& e = getEngine(engine_list.getCurrent()+1);
            item_cost_dpy.setCost(e.getCost());
            item_cost_dpy.setTech(action.getPlanet().getBaseTech(tl_Engine), e.getTech());
            break;
         }
         case 1002: {
            GBeam& b = getBeam(beam_list.getCurrent()+1);
            item_cost_dpy.setCost(b.getCost());
            item_cost_dpy.setTech(action.getPlanet().getBaseTech(tl_Beam), b.getTech());
            break;
         }
         case 1003: {
            GTorpLauncher& t = getTorp(torp_list.getCurrent()+1);
            item_cost_dpy.setCost(t.getCost());
            item_cost_dpy.setTech(action.getPlanet().getBaseTech(tl_Torpedo), t.getTech());
            break;
         }
        }
    }
}

void
WShipBuildScreen::onPlanetChanged(int)
{
    if (UIBaseWidget* p = cards.getFocusedChild()) {
        switch (p->getId()) {
         case 1000:
            updateHullPage();
            hull_list.drawWidget(false);
            break;
         case 1001:
            updateEnginePage();
            engine_list.drawWidget(false);
            break;
         case 1002:
            updateBeamPage();
            beam_list.drawWidget(false);
            break;
         case 1003:
            updateLauncherPage();
            torp_list.drawWidget(false);
            break;
        }
    }
}

/*********************** WShipBuildScreenWithActor ***********************/

WShipBuildScreenWithActor::WShipBuildScreenWithActor(GPlanet& the_planet)
    : WShipBuildScreen(_("Build a Ship"), the_planet)
{
    UIGroup& g = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIButton& ok = h.add(new UIButton(2000, _("Enter - Build"), SDLK_RETURN));
    UIButton& cancel = h.add(new UIButton(2001, _("ESC - Exit"), SDLK_ESCAPE));
    g.add(ok);
    g.add(cancel);
    if (the_planet.getBaseBuildHull()) {
        UIButton& cancel_build = h.add(new UIButton(2002, UIKeyString(_("C - Cancel Build"))));
        cancel_build.sig_fired.add(this, &WShipBuildScreenWithActor::onCancelBuild);
        g.add(cancel_build);
    }
    g.add(h.add(new UISpacer()));
    add(g);

    ok.sig_fired.add(this, &WShipBuildScreenWithActor::onOk);
    cancel.sig_fired.add(this, &WShipBuildScreenWithActor::onCancel);

    pack();
}

WShipBuildScreenWithActor::~WShipBuildScreenWithActor()
{ }

void
WShipBuildScreenWithActor::onOk()
{
    if (GError e = action.getError()) {
        messageBox(format(_("You can not build this ship: %s"), e.getUserError()),
                   _("Build order rejected"));
        return;
    }
    action.doCommit();
    transfer.doCommit();
    stop(1);
}

void
WShipBuildScreenWithActor::onCancel()
{
    action.doCancel();
    transfer.doCancel();
    stop(0);
}

void
WShipBuildScreenWithActor::onCancelBuild()
{
    static const TBuildOrder blank = { 0, 0, 0, 0, 0, 0, 0 };
    action.doCancel();
    transfer.doCancel();
    action.getPlanet().setBaseBuildOrder(blank);
    stop(1);
}
