/**
  *  \file client/dlg-visual.cc
  *  \brief Visual Scanner Dialog
  *
  *  (c) 2001-2009 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/chart.h"
#include "client/dlg-visual.h"
#include "client/scr-sim.h"
#include "client/wtile.h"
#include "cpluslib/format.h"
#include "cpluslib/sigslot.h"
#include "game/ship.h"
#include "game/sim-game.h"
#include "game/sim.h"
#include "game/spec.h"
#include "gfx/palette.h"
#include "resmgr/resid.h"
#include "ui/layout.h"
#include "ui/msgbox.h"
#include "ui/util.h"
#include "ui/widgets.h"
#include "ui/window.h"
#include "util/nls.h"
#include "game/univ.h"
#include "game/planet.h"
#include "client/actions.h"

namespace {
    enum {
        cm_VisSelect,
        cm_VisCancel,
        cm_VisAdd
    };

    class WVisualScanWindow : public UIWindow {
        GObjectList& list;
        int          index;
        SignalConnection conn;

        ObjectHolder h;

        void onScroll();
        void onUpdate();

     public:
        WVisualScanWindow(string_t title, string_t okname, GObjectList& list);
        ~WVisualScanWindow();
        bool handleCommand(const UICommand& cmd, UIBaseWidget& sender);
        bool handleEvent(const UIEvent& e, bool second_pass);
        void drawContent(GfxCanvas& can);
        GObjectId getCurrentObject() const;
        void setIndex(int n);

        UIImageWidget image;   // FIXME?

        Signal0      sig_scrolled;
    };

    class WVisualScanTile : public WTile {
        WVisualScanWindow& window;
        SignalConnection conn;
     public:
        WVisualScanTile(GfxPoint size, WVisualScanWindow& window);
        ~WVisualScanTile();
        GObjectId getCurrentObject()
            { return window.getCurrentObject(); }
    };

    class WVisualScanHeaderTile : public WVisualScanTile {
     public:
        WVisualScanHeaderTile(WVisualScanWindow& window);
        void drawData(GfxCanvas& can);
    };

    class WVisualScanHullInfoTile : public WVisualScanTile {
     public:
        WVisualScanHullInfoTile(WVisualScanWindow& window);
        void drawData(GfxCanvas& can);
        void getLayoutInfo(LayoutInfo& info);
    };

    class WVisualScanShipInfoTile : public WVisualScanTile {
     public:
        WVisualScanShipInfoTile(WVisualScanWindow& window);
        void drawData(GfxCanvas& can);
    };
}

/*************************** WVisualScanWindow ***************************/

WVisualScanWindow::WVisualScanWindow(string_t title, string_t okname, GObjectList& list)
    : UIWindow(title, BLUE_WINDOW, COLOR_GRAY),
      list(list),
      index(1),
      image(105+2, 93+2)
{
    // Widget tree:
    //    Window [VBox]
    //      Group [HBox]      1
    //        Group [VBox]    11
    //          HeaderTile
    //          Group [HBox]  111
    //            Picture
    //            HullInfoTile
    //          ShipInfoTile
    //        Group [VBox]    12
    //          Buttons "Prev", "S", "L", "Next"
    //      Group [HBox]      2
    //        Buttons "OK", "Ins", "H", etc.

    setLayout(UIVBoxLayout::instance5);

    UIGroup& group1   = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& group11  = h.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& group111 = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& group12  = h.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& group2   = h.add(new UIGroup(UIHBoxLayout::instance5));

    add(group1);
    add(group2);

    group1.add(group11);
    group1.add(group12);

    group11.add(h.add(new WVisualScanHeaderTile(*this)));
    group11.add(group111);
    group11.add(h.add(new WVisualScanShipInfoTile(*this)));

    group111.add(image);
    group111.add(h.add(new WVisualScanHullInfoTile(*this)));

    group12.add(h.add(new UIButton("\036", '-')));
    group12.add(h.add(new UISpacer()));
    group12.add(h.add(new UIButton("\037", '+')));

    group2.add(h.add(new UISpacer()));
    group2.add(h.add(new UIButton(cm_VisSelect, okname, SDLK_RETURN)).addCommand());
    group2.add(h.add(new UIButton(cm_VisAdd, _("Add"), SDLK_INSERT)).addCommand());
    group2.add(h.add(new UIButton(cm_VisCancel, _("ESC"), SDLK_ESCAPE)).addCommand());

    onScroll();
}

WVisualScanWindow::~WVisualScanWindow()
{ }

void
WVisualScanWindow::onScroll()
{
    sig_scrolled.raise();

    GObjectId o = getCurrentObject();
    GShip* s;
    if ((s = dynamic_cast<GShip*>(&o.getObject())) != 0) {
        conn = s->sig_changed.add(this, &WVisualScanWindow::onUpdate);
    } else {
        conn = 0;
    }
    onUpdate();
}

void
WVisualScanWindow::onUpdate()
{
    drawWidget(false);

    GObjectId o = getCurrentObject();
    GShip* s;
    if ((s = dynamic_cast<GShip*>(&o.getObject())) != 0) {
        int h = s->getHullId();
        if (h > 0)
            image.setImage(requestPixmap(ResId(res::ship, getHull(h).getPicture(), h)));
        else
            image.setImage(requestPixmap(ResId("nvc")));
    } else {
        image.setImage(0);
    }
}

bool
WVisualScanWindow::handleCommand(const UICommand& cmd, UIBaseWidget& sender)
{
    switch (cmd.code) {
     case cm_VisSelect:
        stop(1);
        return true;
     case cm_VisCancel:
        stop(0);
        return true;
     case cm_VisAdd:
        if (GShip* s = dynamic_cast<GShip*>(&getCurrentObject().getObject()))
            doShipAddToSim(*s, true);
        else if (GPlanet* p = dynamic_cast<GPlanet*>(&getCurrentObject().getObject()))
            doPlanetAddToSim(*p);
        return true;
     default:
        return UIWindow::handleCommand(cmd, sender);
    }
}

bool
WVisualScanWindow::handleEvent(const UIEvent& e, bool second_pass)
{
    int n;
    if (e.type == ev_Key) {
        switch (e.key.code) {
         case '+':
         case SDLK_DOWN:
         case SDLK_PAGEDOWN:
         case ss_Ctrl + '+':
         case ss_Ctrl + SDLK_DOWN:
         case ss_Ctrl + SDLK_PAGEDOWN:
            setSelectedWidget();
            if ((n = list.getNext(index, (e.key.code & ss_Ctrl) != 0)) > 0)
                setIndex(n);
            return true;

         case '-':
         case SDLK_UP:
         case SDLK_PAGEUP:
         case ss_Ctrl + '-':
         case ss_Ctrl + SDLK_UP:
         case ss_Ctrl + SDLK_PAGEUP:
            setSelectedWidget();
            if ((n = list.getPrev(index, (e.key.code & ss_Ctrl) != 0)) > 0)
                setIndex(n);
            return true;

         case '.':
            if (getCurrentObject().isGameObject()) {
                GObject& o = getCurrentObject().getObject();
                o.setIsMarked(!o.isMarked());
                getCurrentObject().getType().getUniverse()->doScreenUpdates();
            }
            return true;
        }
    }
    return UIWindow::handleEvent(e, second_pass);
}

void
WVisualScanWindow::drawContent(GfxCanvas& can)
{
    UIWindow::drawContent(can);

    GObjectId id = getCurrentObject();
    if (id.getObject().isMarked())
        drawColorFrame(can, image.getExtent(), cf_Yellow, true);
    else
        drawColorFrame(can, image.getExtent(), cf_Gray, true);
}

GObjectId
WVisualScanWindow::getCurrentObject() const
{
    if (index > 0 && index <= list.getNumObjects())
        return list.getObjectId(index);
    else
        return GObjectId();
}

void
WVisualScanWindow::setIndex(int n)
{
    if (n > list.getNumObjects())
        n = list.getNumObjects();
    if (n <= 0)
        n = 1;
    if (n != index) {
        index = n;
        onScroll();
    }
}

/**************************** WVisualScanTile ****************************/

WVisualScanTile::WVisualScanTile(GfxPoint size, WVisualScanWindow& window)
    : WTile(size),
      window(window),
      conn(window.sig_scrolled.add((UIBaseWidget*) this, (void (UIBaseWidget::*)()) &UIBaseWidget::drawWidget))
{ }

WVisualScanTile::~WVisualScanTile()
{ }

/************************* WVisualScanHeaderTile *************************/

WVisualScanHeaderTile::WVisualScanHeaderTile(WVisualScanWindow& window)
    : WVisualScanTile(font_sizes[FONT_NORMAL].scaledBy(20, 3).movedBy(0, 3), window)
{ }

void
WVisualScanHeaderTile::drawData(GfxCanvas& can)
{
    setColor(UIColor::tc_Static);

    GObjectId obj = getCurrentObject();
    GShip* s;
    if ((s = dynamic_cast<GShip*>(&obj.getObject())) != 0) {
        /* it's a ship */
        // Line 1: Ship #1: FOOBAR
        //     or: Ship #1 (guessed position!)
        string_t text;
        if (s->getKind() == GShip::GuessedShip)
            text = format(_("Ship #%d (guessed position!)"), s->getId());
        else
            text = format(_("Ship #%d: %s"), s->getId(), s->getName());
        outTextF(can, FONT_NORMAL, 0, 0, extent.w, text);

        // Line 2: (our <ship>)
        //         (our <ship> under X control)
        //         (X <ship>)
        //         (X <ship> under Y control)
        //         (X <ship> under our control)
        // where <ship> is either 'freighter', 'starship', or 'ship'
        enum { Freighter, Starship, Ship } kind;
        if (s->getHullId() > 0) {
            if (s->getKind() == GShip::CurrentShip) {
                if (s->getBeamCount() > 0 || s->getBayCount() > 0 || s->getTorpLauncherCount() > 0)
                    kind = Starship;
                else
                    kind = Freighter;
            } else {
                const GHull& h = getHull(s->getHullId());
                if (h.getMaxBeams() > 0 || h.getNumBays() > 0 || h.getMaxLaunchers() > 0)
                    kind = Starship;
                else
                    kind = Freighter;
            }
        } else {
            kind = Ship;
        }

        if (s->getRealOwner() == getPlayerId()) {
            static const char*const table[] = {
                N_("our freighter"),
                N_("our starship"),
                N_("our ship")
            };
            text = _(table[kind]);
        } else {
            static const char*const table[] = {
                N_("%s freighter"),
                N_("%s starship"),
                N_("%s ship")
            };
            text = format(_(table[kind]), getAdjName(s->getRealOwner()));
        }

        if (s->getRealOwner() != s->getOwner()) {
            if (s->getOwner() == getPlayerId()) {
                text += _(" under our control");
            } else {
                text += format(_(" under %s control"), getAdjName(s->getOwner()));
            }
        }

        setColor(getPlayerColor(s->getOwner()));
        outTextF(can, FONT_NORMAL, 0, font_heights[FONT_NORMAL], extent.w, "(" + text + ")");

        // Line 3: Unknown type
        //         HULL CLASS
        //         Experienced HULL CLASS
        if (s->getHullId() > 0) {
            // FIXME: experience levels
            text = getHull(s->getHullId()).getName();
        } else {
            text = _("Unknown type");
        }
        setColor(UIColor::tc_Static);
        outTextF(can, FONT_NORMAL, 0, 2*font_heights[FONT_NORMAL]+3, extent.w, text);
    } else {
        /* something else */
        outTextF(can, FONT_NORMAL, 0, 0, extent.w, obj.getName());
        outTextF(can, FONT_NORMAL, 0, font_heights[FONT_NORMAL], extent.w, string_t());
        outTextF(can, FONT_NORMAL, 0, 2*font_heights[FONT_NORMAL]+3, extent.w, string_t());
    }
}

/************************ WVisualScannHullInfoTile ***********************/

WVisualScanHullInfoTile::WVisualScanHullInfoTile(WVisualScanWindow& window)
    : WVisualScanTile(font_sizes[FONT_NORMAL].scaledBy(10, 6), window)
{ }

void
WVisualScanHullInfoTile::drawData(GfxCanvas& can)
{
    setColor(UIColor::tc_Static);

    GObjectId obj = getCurrentObject();
    GShip* s;
    int hid;
    if ((s = dynamic_cast<GShip*>(&obj.getObject())) != 0
        && (hid = s->getHullId()) > 0)
    {
        const GHull& h = getHull(hid);
        const int w = extent.w;
        const int hh = font_heights[FONT_NORMAL];

        // Line 1: "Hull mass: nn kt"
        string_t text = format(_("Hull mass: %d kt"), h.getMass());
        outTextF(can, FONT_NORMAL, 0, 0, w, text);

        // Line 2:
        //   Cargo: a/b kt
        //   Max cargo: x kt
        if (s->getKind() == GShip::CurrentShip)
            text = format(_("Cargo: %d/%d kt"), h.getCargo() - s->getFreeCargo(), h.getCargo());
        else
            text = format(_("Max Cargo: %d kt"), h.getCargo());
        outTextF(can, FONT_NORMAL, 0, hh, w, text);

        // Line 3:
        //   Fuel: a/b kt
        //   Max fuel: x kt
        if (s->getKind() == GShip::CurrentShip)
            text = format(_("Fuel: %d/%d kt"), s->getCargo(el_Neutronium), h.getFuel());
        else
            text = format(_("Max Fuel: %d kt"), h.getFuel());
        outTextF(can, FONT_NORMAL, 0, 2*hh, w, text);

        // Line 4:
        //   3xBeamTyp
        //   Max Beams: x
        //   No beams
        // FIXME: handle history
        if (s->getKind() == GShip::CurrentShip) {
            if (s->getBeamCount() > 0)
                text = format("%dx%s", s->getBeamCount(), getBeam(s->getBeamType()).getName());
            else
                text = _("No beams");
        } else {
            if (h.getMaxBeams() > 0)
                text = format(_("Max Beams: %d"), h.getMaxBeams());
            else
                text = _("No beams");
        }
        outTextF(can, FONT_NORMAL, 0, 3*hh, w, text);

        // Line 5:
        //   2xTorpTyp
        //   Max torps: x
        //   No torps
        //   Fighter bays: x
        //   No fighter bays
        if (s->getKind() == GShip::CurrentShip) {
            if (s->getTorpLauncherCount() > 0) {
                text = format("%dx%s", s->getTorpLauncherCount(), getTorp(s->getTorpType()).getName());
            } else if (s->getBayCount() > 0) {
                text = format(_("Fighter bays: %d"), s->getBayCount());
            } else if (h.getNumBays() > 0) {
                text = _("No fighter bays");
            } else if (h.getMaxLaunchers() > 0) {
                text = _("No torps");
            } else {
                text.clear();
            }
        } else {
            if (h.getNumBays() > 0)
                text = format(_("Fighter bays: %d"), h.getNumBays());
            else if (h.getMaxLaunchers() > 0)
                text = format(_("Max torps: %d"), h.getMaxLaunchers());
            else
                text.clear();
        }
        outTextF(can, FONT_NORMAL, 0, 4*hh, w, text);

        // Line 6:
        //   FCode: foo
        if (s->getKind() == GShip::CurrentShip)
            text = format(_("FCode: %s"), s->getFCode());
        else
            text.clear();
        outTextF(can, FONT_NORMAL, 0, 5*hh, w, text);
    } else {
        // Hull not known: completely blank
        getSkin().drawBackground(can, getExtent());
    }
}

void
WVisualScanHullInfoTile::getLayoutInfo(LayoutInfo& info)
{
    WVisualScanTile::getLayoutInfo(info);
    info.flags |= li_FlexibleH;
}

/************************ WVisualScanShipInfoTile ************************/

WVisualScanShipInfoTile::WVisualScanShipInfoTile(WVisualScanWindow& window)
    : WVisualScanTile(font_sizes[FONT_NORMAL].scaledBy(20, 4),
                      window)
{ }

void
WVisualScanShipInfoTile::drawData(GfxCanvas& can)
{
    setColor(UIColor::tc_Static);

    GObjectId obj = getCurrentObject();
    GShip* s;
    int hid;
    if ((s = dynamic_cast<GShip*>(&obj.getObject())) != 0
        && (hid = s->getHullId()) > 0)
    {
        const int w = extent.w;
        const int hh = font_heights[FONT_NORMAL];

        // Line 1:
        //   Mass: x kt        Speed: warp X
        //   Mass: unknown     Not moving
        string_t text;
        if (s->getMass() > 0) {
            text = format(_("Mass: %d kt"), s->getMass());
        } else {
            text = _("Mass: unknown");
        }
        outTextF(can, FONT_NORMAL, 0, 0, w/2, text);

        if (!s->getWarp().isKnown())
            text.clear();
        else if (s->getWarp() > 0)
            text = format(_("Speed: warp %d"), s->getWarp());
        else
            text = _("Not moving");
        outTextF(can, FONT_NORMAL, w/2, 0, w/2, text);

        // Line 2:
        //   Waypoint: foo
        if (s->getKind() == GShip::CurrentShip)
            text = format(_("Waypoint: %s"),
                          getDisplayedTurn().getCurrentUniverse().getLocationName(s->getWaypoint(),
                                                                                  locs_WW));
        else
            text.clear();
        outTextF(can, FONT_NORMAL, 0, hh, w, text);

        // Line 3:
        //   Next turn: foo
        // FIXME: not yet implemented
        text.clear();
        outTextF(can, FONT_NORMAL, 0, 2*hh, w, text);

        // Line 4:
        //   Damage: %d%%
        if (s->getDamage().isKnown())
            text = format(_("Damage: %d%%"), s->getDamage());
        else
            text.clear();
        outTextF(can, FONT_NORMAL, 0, 3*hh, w, text);
    } else {
        // FIXME: we might show heading etc.
        getSkin().drawBackground(can, getExtent());
    }
}

/** Do standard "list ships" dialog.
    \param pt        location to list
    \param flags     flags (ls_XXX)
    \param exclude_ship do not show that ship
    \param title     title line
    \param okname    name of "ok" button
    \param help_id   help page number */
GObjectId
doListShips(GPoint pt, int flags, int exclude_ship,
            string_t title, string_t okname, int /*help_id*/)
{
    /* Initialize list */
    GObjectList l;
    l.addObjectsAt(getDisplayedTurn().getCurrentUniverse(), pt, flags, exclude_ship);
    l.postprocessUser(true);
    int n = l.getNumGameObjects();

    /* List empty? Show message. */
    if (n == 0) {
        string_t msg;
        const GUniverse& univ = getDisplayedTurn().getCurrentUniverse();
        if (exclude_ship != 0
            && (!univ.getShip(exclude_ship).isExistant()
                || univ.getShip(exclude_ship).getPos() != pt))
        {
            /* we were told to exclude a ship, but that ship isn't here */
            exclude_ship = 0;
        }
        
        if (!(flags & ls_IncludeForeignShips)) {
            if (exclude_ship)
                msg = _("There is no other ship of ours at that position.");
            else
                msg = _("There is no ship of ours at that position.");
        } else {
            if (exclude_ship) {
                msg = _("We can't locate another ship at that position.");
            } else {
                msg = _("We can't locate a ship at that position.");
                int pid = univ.getPlanetAt(pt);
                if (pid && !univ.getPlanet(pid).isPlayed())
                    /* this message must start with a space */
                    msg += format(_(" The planet %s may be hiding ships from our sensors."),
                                  univ.getPlanet(pid).getName());
            }
        }
        messageBox(msg, _("Scanner"));
        return GObjectId();
    }

    /* One object only? Bail out early if allowed. */
    if (n == 1 && (flags & ls_EarlyExit)) {
        GObjectList::size_type n = 1;
        while (!l.getObjectId(n).isGameObject())
            ++n;
        if ((flags & ls_AllowForeignShips) || l.getObject(n).isPlayed())
            return l.getObjectId(n);
    }

    /* Do dialog. */
    WVisualScanWindow window(title, okname, l);
    window.pack();
    ui_root->moveWidgetToEdge(window, 0, 0, 5);
    int result = ui_root->runModal(window);
    if (result)
        return window.getCurrentObject();
    else
        return GObjectId();
}
