/**
  *  \file client/scr-mine.cc
  *
  *  (c) 2001-2009 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/screens.h"
#include "client/wtile.h"
#include "cpluslib/format.h"
#include "cpluslib/holder.h"
#include "game/global.h"
#include "game/mine.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/uiglobal.h"
#include "ui/widgets.h"
#include "util/nls.h"
#include "game/planet.h"

namespace {
    class WMinefieldInfoTile : public WObjectTypeObserverTile {
     public:
        WMinefieldInfoTile(GObjectType& type);
        void drawData(GfxCanvas& can);
    };

    class WMinefieldPassageTile : public WObjectTypeObserverTile {
        SignalConnection conn_scroll;
        int move_radius;
        UIButton button_inc, button_dec;
     public:
        WMinefieldPassageTile(GObjectType& type);
        void drawData(GfxCanvas& can);
        void setMoveRadius(int n);
        void onScroll();
        bool handleEvent(const UIEvent& e, bool second_pass);
    };

    class WMinefieldInfoWindow : public UIWindow {
        ObjectHolder h;
        GObjectType& type;
     public:
        WMinefieldInfoWindow(GObjectType& type);

        bool handleEvent(const UIEvent& e, bool second_pass);
    };
}

/** Find planet associated with a minefield.
    \return planet Id, or 0 */
static int
getAssociatedPlanet(const GGameTurn& turn, const GMinefield& mf)
{
    // chartdlg.pas, CMinefieldView.SetMinefield

    int pid  = 0;
    long dist = 0;
    for (int i = 1; i <= NUM_PLANETS; ++i) {
        const GPlanet& p = turn.getPlanet(i);
        if (p.isExistant()) {
            int thisown = p.getOwner();
            bool maybe;

            /* The planet is a possible controlling planet if...
               - we know it has the same owner as the minefield
               - we don't know the planet's owner, and don't have
                 full data for the minfield owner's race */
            if (thisown > 0)
                maybe = (thisown == mf.getOwner());
            else
                maybe = turn.isData(mf.getOwner());

            /* Choose closest planet */
            if (maybe) {
                long ndist = distanceSquared(p.getPos(), mf.getPos());
                if (pid == 0 || ndist < dist) {
                    pid  = i;
                    dist = ndist;
                }
            }
        }
    }

    return pid;
}

/************************* WObjectTypeHeaderTile *************************/

WObjectTypeHeaderTile::WObjectTypeHeaderTile(GObjectType& type)
    : WObjectTypeObserverTile(font_sizes[FONT_NORMAL].scaledBy(20, 1), type),
      button_next(0, "\037", SDLK_PAGEUP),
      button_prev(0, "\036", SDLK_PAGEDOWN)
{
    static const int h = font_heights[FONT_NORMAL];
    setLayout(RightJustified, 0, false);
    button_prev.setState(UIButton::bf_SmallSize, true);
    button_prev.addCall(this, &WObjectTypeHeaderTile::onPrev);
    button_next.setState(UIButton::bf_SmallSize, true);
    button_next.addCall(this, &WObjectTypeHeaderTile::onNext);
    addControl(GfxRect(0, 0, h-1, h-1), button_next);
    addControl(GfxRect(h, 0, h-1, h-1), button_prev);
}

void
WObjectTypeHeaderTile::drawData(GfxCanvas& can)
{
    string_t name;
    if (GObject* o = getObject()) {
        name = o->getName();
        name += ":";
    }

    setColor(UIColor::tc_Static);
    outTextF(can, FONT_NORMAL, 0, 0, extent.w - 2*font_heights[FONT_NORMAL], name);
}

void
WObjectTypeHeaderTile::getLayoutInfo(LayoutInfo& info)
{
    WObjectTypeObserverTile::getLayoutInfo(info);
    info.flags |= li_FlexibleH;
}

void
WObjectTypeHeaderTile::onNext()
{
    GObjectType& t = getType();
    t.setCurrent(t.getNext(t.getCurrent(), false));
}

void
WObjectTypeHeaderTile::onPrev()
{
    GObjectType& t = getType();
    t.setCurrent(t.getPrev(t.getCurrent(), false));
}

/*************************** WMinefieldInfoTile **************************/

WMinefieldInfoTile::WMinefieldInfoTile(GObjectType& type)
    : WObjectTypeObserverTile(font_sizes[FONT_NORMAL].scaledBy(28, 7), type)
{ }

void
WMinefieldInfoTile::drawData(GfxCanvas& can)
{
    // |----12em-----|------16em----------------|
    // Owner:         Fed
    // Size:          32 ly radius
    //                999 units
    // After decay:   900 units (30 ly)
    // Last info:     current turn
    // Controlled by: Wolf 359
    //                Fed planet

    GfxRect pos  = getInnerSize(0, 0);
    const int dy = font_heights[FONT_NORMAL];
    const int x0 = 0;
    const int x1 = font_ems[FONT_NORMAL]*12;
    const int w  = pos.w - x1;
    int y = 0;

    GMinefield* mf = dynamic_cast<GMinefield*>(getObject());
    if (!mf)
        return;

    setColor(UIColor::tc_Static);

    // Line 1:
    outText (can, FONT_NORMAL, x0, y, _("Owner:"));
    outTextF(can, FONT_NORMAL, x1, y, w, getShortName(mf->getOwner()));
    y += dy;

    // Line 2+3:
    outText (can, FONT_NORMAL, x0, y, _("Size:"));
    outTextF(can, FONT_NORMAL, x1, y, w, format(_("%d ly radius"), mf->getRadius()));
    y += dy;
    outTextF(can, FONT_NORMAL, x1, y, w, format(_("%d units"), mf->getUnits()));
    y += dy;

    // Line 4:
    int32 uad = mf->getUnitsAfterDecay(mf->getUnits());
    outText (can, FONT_NORMAL, x0, y, _("After decay:"));
    outTextF(can, FONT_NORMAL, x1, y, w, format(_("%d units (%d ly)"), uad, mf->getRadiusFromUnits(uad)));
    y += dy;

    // Line 5:
    int age = getDisplayedTurn().getTurnNumber() - mf->getTurnLastSeen();
    outText (can, FONT_NORMAL, x0, y, _("Last info:"));
    if (age == 0)
        outTextF(can, FONT_NORMAL, x1, y, w, _("this turn"));
    else if (age == 1)
        outTextF(can, FONT_NORMAL, x1, y, w, _("last turn"));
    else if (age > 0)
        outTextF(can, FONT_NORMAL, x1, y, w, format(_("%d turns ago"), age));
    else
        outTextF(can, FONT_NORMAL, x1, y, w, format(_("Turn %d"), mf->getTurnLastSeen()));
    y += dy;

    // Lines 6+7:
    int pid = getAssociatedPlanet(getDisplayedTurn(), *mf);
    outText (can, FONT_NORMAL, x0, y, _("Controlled by:"));
    if (pid == 0) {
        outTextF(can, FONT_NORMAL, x1, y, w, _("unknown"));
        y += dy;
        outTextF(can, FONT_NORMAL, x1, y, w, string_t());
    } else {
        const GPlanet& p = getDisplayedTurn().getPlanet(pid);
        outTextF(can, FONT_NORMAL, x1, y, w, p.getName());
        y += dy;
        if (p.getOwner() < 0)
            outTextF(can, FONT_NORMAL, x1, y, w, _("a planet with unknown owner"));
        else if (p.getOwner() == 0)
            outTextF(can, FONT_NORMAL, x1, y, w, _("unowned planet"));
        else if (p.getOwner() == getPlayerId())
            outTextF(can, FONT_NORMAL, x1, y, w, _("our planet"));
        else
            outTextF(can, FONT_NORMAL, x1, y, w, getShortName(p.getOwner()));
    }
    y += dy;
}

/************************* WMinefieldPassageTile *************************/

WMinefieldPassageTile::WMinefieldPassageTile(GObjectType& type)
    : WObjectTypeObserverTile(font_sizes[FONT_NORMAL].scaledBy(28, 2), type),
      button_inc("+", '+'),
      button_dec("-", '-')
{
    conn_scroll = type.sig_pos_changed.add(this, &WMinefieldPassageTile::onScroll);
    onScroll();

    const int h = font_heights[FONT_NORMAL];
    setLayout(RightJustified, 0, true);
    button_inc.setState(UIButton::bf_SmallSize, true);
    button_dec.setState(UIButton::bf_SmallSize, true);
    addControl(GfxRect(0, 0, h-1, h-1), button_inc);
    addControl(GfxRect(0, h, h-1, h-1), button_dec);
}

void
WMinefieldPassageTile::drawData(GfxCanvas& can)
{
    // |----12em-----|------16em----------------|
    // Safe passage:  32 ly - 72.5%           [+]
    //                32 ly - 85.2% (cloaked) [-]

    GfxRect pos  = getInnerSize(0, 0);
    const int dy = font_heights[FONT_NORMAL];
    const int x0 = 0;
    const int x1 = font_ems[FONT_NORMAL]*12;
    const int w  = pos.w - x1;
    int y = 0;

    GMinefield* mf = dynamic_cast<GMinefield*>(getObject());
    if (!mf)
        return;

    setColor(UIColor::tc_Static);

    // Lines 1+2:
    outText (can, FONT_NORMAL, x0, y, _("Successful passage:"));
    outTextF(can, FONT_NORMAL, x1, y, w,
             format(_("%d ly - %.1f%%"),           move_radius, 100.0 * mf->getPassRate(move_radius, false)));
    y += dy;
    outTextF(can, FONT_NORMAL, x1, y, w,
             format(_("%d ly - %.1f%% (cloaked)"), move_radius, 100.0 * mf->getPassRate(move_radius, true)));
    y += dy;
}

void
WMinefieldPassageTile::setMoveRadius(int n)
{
    // Range check
    if (n < 0)
        n = 0;
    if (n > 1000 /* FIXME */)
        n = 1000;

    // Set
    if (n != move_radius) {
        move_radius = n;
        drawWidget();
    }
}

void
WMinefieldPassageTile::onScroll()
{
    if (GMinefield* mf = dynamic_cast<GMinefield*>(getObject()))
        setMoveRadius(mf->getRadius());
}

bool
WMinefieldPassageTile::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        switch (e.key.code) {
         case '+':
            setMoveRadius(move_radius+10);
            return true;
         case '-':
            setMoveRadius(move_radius-10);
            return true;
         case '+' + ss_Shift:
            setMoveRadius(move_radius+1);
            return true;
         case '-' + ss_Shift:
            setMoveRadius(move_radius-1);
            return true;
        }
    }
    return WObjectTypeObserverTile::handleEvent(e, second_pass);
}

/************************** WMinefieldInfoWindow *************************/

WMinefieldInfoWindow::WMinefieldInfoWindow(GObjectType& type)
    : UIWindow(_("Minefield Information"), BLUE_WINDOW, COLOR_GRAY),
      type(type)
{
    // VBox
    //   WObjectTypeHeaderTile
    //   WMinefieldInfoTile
    //   WMinefieldPassageTile
    //   HBox
    //     Close GoTo Planet Del <--> H
    // Difference to PCC 1.x: Up/Down are in WObjectTypeHeaderTile
    // FIXME: where are S and I?
    setLayout(UIVBoxLayout::instance5);

    UIGroup& g1 = h.add(new UIGroup(UIHBoxLayout::instance5));

    g1.add(h.add(new UIButton(_("Close"), SDLK_ESCAPE)));
    g1.add(h.add(new UIButton(_("Go To"), SDLK_RETURN)));
    g1.add(h.add(new UISpacer()));

    add(h.add(new WObjectTypeHeaderTile(type)));
    add(h.add(new WMinefieldInfoTile(type)));
    add(h.add(new WMinefieldPassageTile(type)));
    add(g1);

    pack();
}

bool
WMinefieldInfoWindow::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        switch (e.key.code) {
         case SDLK_UP:
         case SDLK_PAGEUP:
         case vk_WheelUp:
            type.setCurrent(type.getPrev(type.getCurrent(), false));
            return true;
         case SDLK_DOWN:
         case SDLK_PAGEDOWN:
         case vk_WheelDown:
            type.setCurrent(type.getNext(type.getCurrent(), false));
            return true;
         case SDLK_RETURN:
            stop(1);
            return true;
         case SDLK_ESCAPE:
            stop(0);
            return true;
        }
    }
    return UIWindow::handleEvent(e, second_pass);
}

WScreenId
doMinefieldScreen()
{
    if (global_minefield_type.getCount() == 0)
        return WScreenId();

    WMinefieldInfoWindow win(global_minefield_type);
    ui_root->centerWidget(win);
    int result = ui_root->runModal(win);
    if (result == 1) {
        // "Go to"
        return WScreenId(global_minefield_type.getMinefield().getPos());
    } else {
        // "Close"
        return WScreenId();
    }
}

/** Find nearest minefield. Returns 0 if none found. */
int
findNearestMinefield(GPoint pt)
{
    int found = 0;
    long dist2 = 0;
    bool inside = false;
    const GMinefieldContainer& mfs = getDisplayedTurn().getCurrentUniverse().getMinefields();
    for (GMinefieldContainer::const_iterator i = mfs.begin(); i != mfs.end(); ++i) {
        long ndist2 = distanceSquared(i->getPos(), pt);
        bool ninside = (ndist2 <= i->getUnits());
        /* Pick this minefield, if
           - it's the first one
           - it's closer than our previous choice
           - we're inside it but not inside the previous one */
        if (found <= 0
            || ndist2 < dist2
            || (ninside && !inside))
        {
            found  = i->getId();
            dist2  = ndist2;
            inside = ninside;
        }
    }
    return found;
}
