/**
  *  \file client/screens.cc
  *
  *  (c) 2001-2009 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/screens.h"
#include "cpluslib/assert.h"
#include "game/ship.h"
#include "game/planet.h"
#include "game/global.h"
#include "util/nls.h"
#include "cpluslib/format.h"

/** Create blank screen Id.
    \post !isValid */
WScreenId::WScreenId()
    : screen_id(NullScreen), x(0), y(0)
{ }

/** Create screen Id for an object.
    \param screen_id Screen identifier
    \param id        Selected object on that screen. Must be valid (not checked!)
                     or zero. */
WScreenId::WScreenId(ScreenId screen_id, int id)
    : screen_id(screen_id), x(id), y(0)
{
    // FIXME: add this when all code is converted
    // ASSERT(screen_id != Starchart);
}

/** Create screen Id for map location.
    \param pt Map location */
WScreenId::WScreenId(GPoint pt)
    : screen_id(Starchart), x(pt.x), y(pt.y)
{ }

/** Get object Id. */
int
WScreenId::getId() const
{
    ASSERT(screen_id != Starchart);
    ASSERT(screen_id != NullScreen);
    return x;
}

/** Get starchart position.
    \pre getScreenId() == Starchart */
GPoint
WScreenId::getPos() const
{
    ASSERT(screen_id == Starchart);
    return GPoint(x, y);
}

bool
WScreenId::isValid() const
{
    switch (screen_id) {
     case RaceScreen:
     case Starchart:
        return true;

     case ShipScreen:
     case ShipTaskScreen:
        return global_ship_type.isExistant(x);

     case PlanetScreen:
     case PlanetTaskScreen:
        return global_planet_type.isExistant(x);

     case BaseScreen:
     case BaseTaskScreen:
        return global_base_type.isExistant(x);

     default:
        return false;
    }
}

string_t
WScreenId::toString()
{
    switch (screen_id) {
     case RaceScreen:
        return _("Race Screen");
     case Starchart:
        return format(_("Starchart (%d,%d)"), x, y);
     case ShipScreen:
        return format(_("Ship: %s"), getDisplayedTurn().getShip(x).getFullName());
     case PlanetScreen:
        return format(_("Planet: %s"), getDisplayedTurn().getPlanet(x).getFullName());
     case BaseScreen:
        return format(_("Starbase: %s"), getDisplayedTurn().getPlanet(x).getFullName());
     case ShipTaskScreen:
        return format(_("Ship Task: %s"), getDisplayedTurn().getShip(x).getFullName());
     case PlanetTaskScreen:
        return format(_("Planet Task: %s"), getDisplayedTurn().getPlanet(x).getFullName());
     case BaseTaskScreen:
        return format(_("Starbase Task: %s"), getDisplayedTurn().getPlanet(x).getFullName());
     default:
        return format("<%d,%d,%d>", int(screen_id), x, y);
    }
}

/** Check for valid screen Id.
    \return true iff n can be cast to ScreenId */
bool
WScreenId::isValidScreenId(int n)
{
    return n == ShipScreen
        || n == PlanetScreen
        || n == BaseScreen
        || n == Starchart
        || n == ShipTaskScreen
        || n == PlanetTaskScreen
        || n == BaseTaskScreen;
}

/****************************** Applications *****************************/

WScreenId
doSelectObjectForScreen(WScreenId::ScreenId screen_id)
{
    switch (screen_id) {
     case WScreenId::ShipScreen:
     case WScreenId::ShipTaskScreen:
        if (int id = chooseShip())
            return WScreenId(screen_id, id);
        break;
     case WScreenId::PlanetScreen:
     case WScreenId::PlanetTaskScreen:
        if (int id = choosePlanet())
            return WScreenId(screen_id, id);
        break;
     case WScreenId::BaseScreen:
     case WScreenId::BaseTaskScreen:
        if (int id = chooseBase())
            return WScreenId(screen_id, id);
        break;
     case WScreenId::NullScreen:
     case WScreenId::RaceScreen:
     case WScreenId::Starchart:
        break;
    }
    return WScreenId();
}
