/**
  *  \file client/wscore.cc
  *  \brief Score Display - Utilities
  *
  *  (c) 2001-2009 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/wscore.h"
#include "cpluslib/assert.h"

/** Add component to this score. A score can consist of up to NUM components,
    each weighed by an integral weight (e.g. 120=base in Tim-score).
    \param column Column from score table to use
    \param factor Weight */
WScore&
WScore::add(GStatIndex column, int factor)
{
    ASSERT(num_parts < NUM);
    columns[num_parts] = column;
    factors[num_parts] = factor;
    ++num_parts;
    return *this;
}

/** Compute this score value.
    \param record Record to look at
    \param player Player to compute score for
    \return score, -1 if unknown */
int32
WScore::get(const GStatRecord& record, int player) const
{
    int32 total = 0;
    bool hadone = false;
    for (int i = 0; i < num_parts; ++i) {
        int32 me = record(columns[i], player);
        if (me >= 0) {
            hadone = true;
            total += factors[i] * me;
        }
    }
    if (hadone)
        return total;
    else
        return -1;
}

/** Get total amount of a score this turn. This is the sum of all players' scores.
    \param record Record to look up
    \return total score, -1 if no value unknown */
int32
WScore::getTotal(const GStatRecord& record) const
{
    int32 total = 0;
    bool hadone = false;
    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        int32 me = get(record, i);
        if (me >= 0) {
            hadone = true;
            total += me;
        }
    }
    if (hadone)
        return total;
    else
        return -1;
}

/** Get maximum of a score. Returns the maximum value appearing in the recordings.
    \param file Score file to look at
    \param first First record to look at
    \param last One-past-last record to look at
    \return maximum, -1 if no value known */
int32
WScore::getMax(const GStatFile& file, uint32 first, uint32 last) const
{
    int32 max = -1;
    if (num_parts > 0) {
        /* The num_parts check isn't strictly necessary but saves a little
           work when constructing a WScoreChart */
        for (uint32 i = first; i < last; ++i) {
            const GStatRecord& r = file.getRecordByIndex(i);
            for (int pl = 1; pl <= NUM_PLAYERS; ++pl) {
                int32 val = get(r, pl);
                if (val > max)
                max = val;
            }
        }
    }
    return max;
}
