/**
  *  \file client/wscorechart.cc
  *
  *  (c) 2001-2009 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/wscore.h"
#include "ui/uiglobal.h"
#include "util/misc.h"
#include "game/stat.h"
#include "util/nls.h"
#include "gfx/font.h"
#include "gfx/palette.h"
#include "client/wracelist.h"

WScoreChart::WScoreChart(const GStatFile& file, uint32 first, uint32 last)
    : file(file),
      first(first),
      last(last),
      score("?"),
      current_player(0),
      left_margin(0),
      right_margin(0),
      bottom_margin(0),
      max_score(0),
      victory_level(-1),
      victory_turns_needed(0)
{
    victory_turns.setAll(0);
    recomputeGrid();
}

void
WScoreChart::setScore(WScore score)
{
    this->score = score;
    recomputeGrid();
    drawWidget();
}

void
WScoreChart::setCurrentPlayer(int player)
{
    if (current_player != player) {
        current_player = player;
        drawWidget();
    }
}

void
WScoreChart::getLayoutInfo(LayoutInfo& info)
{
    info.min_size  = GfxPoint(left_margin + right_margin + 200, 5*bottom_margin);
    info.pref_size = GfxPoint(left_margin + right_margin + 400, 10*bottom_margin);
    info.flags     = li_Flexible;
}

void
WScoreChart::drawContent(GfxCanvas& can)
{
    /* can we draw? */
    if (extent.w < left_margin + right_margin + 2 || extent.h < bottom_margin)
        return;

    /* figure out position of grid */
    const int x1 = extent.x + left_margin;
    const int x2 = extent.x + extent.w - right_margin;
    const int y1 = extent.y;
    const int y2 = extent.y + extent.h - bottom_margin;

    /* draw grid */
    UISkin& s = getSkin();
    s.startDraw();
    s.drawBackground(can, extent);
    s.setColor(UIColor::tc_Static);

    drawVLine(can, s, x1, y1, y2);       /* Y (score) axis */
    drawHLine(can, s, x1, y2, x2);       /* X (time) axis */
    drawHLine(can, s, x1-1, y1+1, x1+1); /* Y arrowhead */
    drawHLine(can, s, x1-2, y1+2, x1+2);
    drawVLine(can, s, x2-1, y2-1, y2+1); /* X arrowhead */
    drawVLine(can, s, x2-2, y2-2, y2+2);

    s.getParameters().just_x = 2;
    s.outText(can, FONT_NORMAL, x1-2, y1, numToString(max_score));
    s.outText(can, FONT_NORMAL, x1-2, y2, numToString(file.getRecordByIndex(first).getTurnNumber()));
    s.getParameters().just_x = 0;
    s.outText(can, FONT_NORMAL, x2+2, y2, numToString(file.getRecordByIndex(last-1).getTurnNumber()));
    s.getParameters().just_y = 2;
    s.outText(can, FONT_NORMAL, x2+2, y2, _("Turn"));
    s.getParameters().just_x = 2;
    s.outText(can, FONT_NORMAL, x1-2, y2, _("Turn"));

    /* Win limit */
    if (victory_level > 0) {
        s.getParameters().setLineStyle(0xAA);

        int height = extent.h - bottom_margin;
        int y      = y2 - height*victory_level / max_score;
        drawHLine(can, s, x1, y, x2);
    }

    /* draw chart */
    for (int i = 1; i <= NUM_PLAYERS; ++i)
        if (i != current_player)
            drawPlayerChart(can, i, false);
    if (current_player >= 1 && current_player <= NUM_PLAYERS)
        drawPlayerChart(can, current_player, true);

    can.update(extent);
}

void
WScoreChart::recomputeGrid()
{
    /* Compute victory information */
    findVictoryCondition();

    /* Figure out maximum score */
    max_score = score.getMax(file, first, last);
    if (max_score <= 0)
        max_score = 1;

    /* Compute grid ranges */
    int twidth = fonts[FONT_NORMAL]->getTextWidth(_("Turn"));
    int mwidth = fonts[FONT_NORMAL]->getTextWidth(numToString(max_score));

    left_margin   = (twidth > mwidth ? twidth : mwidth) + 2;
    right_margin  = twidth + 2;
    bottom_margin = font_heights[FONT_NORMAL];
}

void
WScoreChart::drawPlayerChart(GfxCanvas& can, int player, bool bold)
{
    GfxParameters parm;

    int prev_x = 0, prev_y = 0;
    int first_turn = file.getRecordByIndex(first).getTurnNumber();
    int turns = file.getRecordByIndex(last-1).getTurnNumber() - first_turn;

    int width  = extent.w - right_margin - left_margin - 2;
    int height = extent.h - bottom_margin;
    int x0     = extent.x + left_margin;
    int y0     = extent.y + height;

    bool drawing = false;

    parm.color          = standard_colors[WRaceList::colors[player-1]];
    parm.line_thickness = bold ? 3 : 1;

    for (uint32 i = first; i < last; ++i) {
        const GStatRecord& r = file.getRecordByIndex(i);

        int32 value = score.get(r, player);
        if (value >= 0) {
            int x = x0 + width*(r.getTurnNumber()-first_turn) / turns;
            int y = y0 - height*value / max_score;
            if (drawing)
                drawLine(can, parm, prev_x, prev_y, x, y);
            can.drawPixel(x-1, y, parm.color, GFX_ALPHA_OPAQUE);
            can.drawPixel(x+1, y, parm.color, GFX_ALPHA_OPAQUE);
            can.drawPixel(x, y-1, parm.color, GFX_ALPHA_OPAQUE);
            can.drawPixel(x, y+1, parm.color, GFX_ALPHA_OPAQUE);
            prev_x = x;
            prev_y = y;
            drawing = true;
        } else {
            drawing = false;
        }
    }
}

void
WScoreChart::findVictoryCondition()
{
    victory_level = -1;
    victory_turns_needed = 0;
    victory_turns.setAll(0);

    /* Only single-part charts can have a sensible victory condition. */
    if (score.getNumParts() != 1 || score.getFactor(0) != 1)
        return;

    /* Do we have a description for this score? */
    const TScoreDescription* pdesc = file.getDescriptionForScoreId(file.getScoreId(score.getPart(0)));
    if (pdesc == 0)
        return;

    /* Does this description specify a victory condition? */
    if (pdesc->win_limit <= 0)
        return;

    victory_level = pdesc->win_limit;
    victory_turns_needed = pdesc->turn_limit > 0 ? pdesc->turn_limit : 1;

    /* Find victory turns, for each player. */
    for (uint32 index = file.getNumTurns(); index > 0; --index) {
        const GStatRecord& record = file.getRecordByIndex(index-1);
        bool any_player_over_limit = false;
        for (int i = 1; i <= NUM_PLAYERS; ++i) {
            if (score.get(record, i) >= victory_level) {
                victory_turns[i] = record.getTurnNumber();
                any_player_over_limit = true;
            }
        }
        if (!any_player_over_limit)
            break;
    }
}

bool
WScoreChart::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        if (e.key.code == ss_Ctrl + SDLK_LEFT) {
            setSelectedWidget();
            int n = current_player - 1;
            if (n <= 0)
                n = NUM_PLAYERS;
            setCurrentPlayer(n);
            return true;
        } else if (e.key.code == ss_Ctrl + SDLK_RIGHT) {
            setSelectedWidget();
            int n = current_player + 1;
            if (n > NUM_PLAYERS)
                n = 1;
            setCurrentPlayer(n);
            return true;
        } else {
            return UIWidget::handleEvent(e, second_pass);
        }
    } else {
        return UIWidget::handleEvent(e, second_pass);
    }
}
