/**
  *  \file client/wscoretable.cc
  *  \brief Score Display - Table
  *
  *  (c) 2001-2009 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/wscore.h"
#include "cpluslib/format.h"
#include "game/host.h"
#include "game/player.h"
#include "game/stat.h"
#include "ui/skin.h"
#include "ui/uiglobal.h"
#include "util/misc.h"
#include "util/nls.h"


/** Names for WScoreTable modes. May contain placeholder for compare turn. */
static const char*const mode_names[] = {
    N_("Normal scoreboard"),
    N_("Differences to previous turn"),
    N_("Differences to turn %d"),
    N_("Percentages"),
    N_("Ratio of total")
};

/** Constructor. You must call addColumn() to add columns. The widget should not
    be used without columns.
    \param file Score file to display. Must not be empty. */
WScoreTable::WScoreTable(const GStatFile& file)
    : file(file),
      columns(),
      current_player(0),
      mode(Normal),
      sort_order(0),
      label_width(font_ems[FONT_NORMAL]*10),
      column_width(font_ems[FONT_NORMAL]*6),
      header_height(font_heights[FONT_TITLE] + font_heights[FONT_NORMAL]/4),
      line_height(font_heights[FONT_NORMAL]  + font_heights[FONT_NORMAL]/4),
      footer_height(font_heights[FONT_NORMAL]),
      mouse_down(false),
      current_record(file.getNumTurns() - 1),
      compare_record(0)
{
    updateSortOrder();
}

/** Add a column. Should only be called while widget is invisible. */
void
WScoreTable::addColumn(const WScore& score)
{
    columns.push_back(score);
    if (int(columns.size())-1 == sort_order)
        updateSortOrder();
}

/** Add default columns. These are the columns provided by PCC 1.x. On every
    modification of the score file, we make sure these columns are actually
    present in the data file, but otherwise we allow display of any subset
    thereof.
    \see addColumn() */
void
WScoreTable::addDefaultColumns()
{
    GStatIndex pi = file.getIndexForScore(ScoreId_Planets);
    GStatIndex bi = file.getIndexForScore(ScoreId_Bases);
    GStatIndex fi = file.getIndexForScore(ScoreId_Freighters);
    GStatIndex ci = file.getIndexForScore(ScoreId_Capital);
    GStatIndex qi = file.getIndexForScore(ScoreId_BuildPoints);

    if (pi != file.nil && bi != file.nil && fi != file.nil && ci != file.nil)
        addColumn(WScore(_("Score")).add(pi, 10).add(bi, 120).add(fi, 1).add(ci, 10));
    if (pi != file.nil)
        addColumn(WScore(_("Planets")).add(pi, 1));
    if (fi != file.nil)
        addColumn(WScore(_("Fr.")).add(fi, 1));
    if (ci != file.nil)
        addColumn(WScore(_("Cap.")).add(ci, 1));
    if (bi != file.nil)
        addColumn(WScore(_("Bases")).add(bi, 1));

    if (qi != file.nil) {
        if (host.isPBPGame())
            addColumn(WScore(_("PBPs")).add(qi, 1));
        else
            addColumn(WScore(_("PAL")).add(qi, 1));
    }
}

/** Get number of columns. */
int
WScoreTable::getNumColumns() const
{
    return columns.size();
}

/** Change sort order.
    \param order Column to sort by; -1 to sort by player number */
void
WScoreTable::setSortOrder(const int order)
{
    if (sort_order != order) {
        sort_order = order;
        updateSortOrder();
        drawWidget();
        sig_config_changed.raise();
    }
}

/** Set display mode. When activating DiffToParticular mode, the record to
    compare with is set to the current record.
    \param mode Mode to use */
void
WScoreTable::setMode(Mode mode)
{
    if (this->mode != mode || mode == DiffToParticular) {
        this->mode = mode;
        if (mode == DiffToParticular)
            compare_record = current_record;
        drawWidget();
        sig_config_changed.raise();
    }
}

/** Set current player. This player is highlighted in the table.
    \param player Player number. Unless Percentages mode is active, can also
                  be 0 to disable highlighting */
void
WScoreTable::setCurrentPlayer(int player)
{
    this->current_player = player;
    drawWidget();
}

/** Set current record, by number.
    \param record Record to display */
void
WScoreTable::setCurrentRecord(uint32 record)
{
    if (current_record != record) {
        current_record = record;
        updateSortOrder();
        drawWidget();
        sig_turn_changed.raise();
    }
}

void
WScoreTable::getLayoutInfo(LayoutInfo& info)
{
    info.min_size = info.pref_size = GfxPoint(label_width + getNumColumns()*column_width,
                                              header_height + NUM_PLAYERS*line_height + footer_height);
    info.flags = 0;
}

void
WScoreTable::drawContent(GfxCanvas& can)
{
    /* figure out what to draw */
    const GStatRecord& me = file.getRecordByIndex(current_record);
    const GStatRecord& other = file.getRecordByIndex(mode == DiffToParticular ? compare_record
                                                     : mode == DiffToPrevious && current_record > 0 ? current_record-1
                                                     : 0);

    std::vector<int32> totals(columns.size(), int32(0));

    UISkin& s = getSkin();
    s.startDraw();

    /* Draw heading */
    s.getParameters().just_x = 0;
    s.getParameters().just_y = 0;
    s.setColor(UIColor::tc_Static);
    s.outTextF(can, FONT_TITLE, extent.x, extent.y, label_width, format(_("Turn %d"), me.getTurnNumber()));
    s.getParameters().just_x = 1;
    for (unsigned i = 0; i < columns.size(); ++i) {
        s.outText(can, FONT_TITLE, extent.x + label_width + i*column_width + column_width/2, extent.y, columns[i].getName());
    }

    /* Draw content */
    for (int i = 0; i < NUM_PLAYERS; ++i) {
        const int pl = players[i];
        int       x = extent.x;
        const int y = extent.y + header_height + i*line_height;

        /* Pick color */
        if (pl == current_player)
            s.setColor(UIColor::tc_Green);
        else if (columns.size() > 0 && columns[0].get(me, pl) > 0)
            s.setColor(UIColor::tc_Static);
        else
            s.setColor(UIColor::tc_Faded);

        /* Draw race name */
        s.getParameters().just_x = 0;
        s.outTextF(can, FONT_NORMAL, x, y, label_width, getShortName(pl));
        x += label_width;

        /* Draw data */
        s.getParameters().just_x = 2;
        for (unsigned col = 0; col < columns.size(); ++col) {
            const int32 mescore = columns[col].get(me, pl);
            if (mescore >= 0) {
                totals[col] += mescore;
                switch (mode) {
                 case Normal: {
                    s.outTextF(can, FONT_NORMAL, x + column_width, y, column_width, numToString(mescore));
                    break;
                 }

                 case DiffToPrevious:
                 case DiffToParticular: {
                    const int32 otherscore = columns[col].get(other, pl);
                    if (otherscore >= 0) {
                        // We know the difference
                        const int32 diff = mescore - otherscore;
                        if (diff == 0)
                            s.outTextF(can, FONT_NORMAL, x + column_width, y, column_width, "0");
                        else
                            s.outTextF(can, FONT_NORMAL, x + column_width, y, column_width, format("%+d", diff));
                        totals[col] -= otherscore;
                    } else {
                        // Previous value not known
                        s.outTextF(can, FONT_NORMAL, x + column_width, y, column_width, numToString(mescore));
                    }
                    break;
                 }

                 case Percentages:
                 case RatioOfTotal: {
                    const int32 otherscore = (mode == RatioOfTotal
                                              ? columns[col].getTotal(me)
                                              : (current_player > 0
                                                 ? columns[col].get(me, current_player)
                                                 : 0));
                    if (otherscore > 0) {
                        // Reference is known and nonzero
                        const int32 percent = 100*mescore / otherscore;
                        if (percent > 9999)
                            s.outTextF(can, FONT_NORMAL, x + column_width, y, column_width, "++++");
                        else
                            s.outTextF(can, FONT_NORMAL, x + column_width, y, column_width, format("%d%%", percent));
                    } else {
                        // Reference value not known or zero
                        s.outTextF(can, FONT_NORMAL, x + column_width, y, column_width, numToString(mescore));
                    }
                    break;
                 }
                }
            } else {
                s.outTextF(can, FONT_NORMAL, x + column_width, y, column_width, "-");
            }
            x += column_width;
        }
    }

    /* Draw totals */
    const int y = extent.y + header_height + NUM_PLAYERS*line_height;
    int       x = extent.x;
    s.getParameters().just_x = 0;
    s.setColor(UIColor::tc_Static);

    s.outTextF(can, FONT_NORMAL, x, y, label_width, _(">> Totals"));
    x += label_width;

    for (unsigned int col = 0; col < totals.size(); ++col) {
        s.getParameters().just_x = 2;
        if (mode == DiffToParticular || mode == DiffToPrevious) {
            if (totals[col] == 0)
                s.outTextF(can, FONT_NORMAL, x + column_width, y, column_width, "0");
            else
                s.outTextF(can, FONT_NORMAL, x + column_width, y, column_width, format("%+d", totals[col]));
        } else {
            s.outTextF(can, FONT_NORMAL, x + column_width, y, column_width, numToString(totals[col]));
        }
        x += column_width;
    }

    can.update(extent);
}

bool
WScoreTable::handleEvent(const UIEvent& evt, bool second_pass)
{
    if (evt.type == ev_Key) {
        switch (evt.key.code) {
         case SDLK_PAGEUP:
         case SDLK_UP:
         case '-':
            setSelectedWidget();
            previousTurn();
            return true;

         case SDLK_PAGEDOWN:
         case SDLK_DOWN:
         case '+':
            setSelectedWidget();
            nextTurn();
            return true;

         case ss_Ctrl + SDLK_UP:
            setSelectedWidget();
            previousPlayer();
            return true;

         case ss_Ctrl + SDLK_DOWN:
            setSelectedWidget();
            nextPlayer();
            return true;

         case ss_Ctrl + 'u':
            setSelectedWidget();
            setSortOrder(-1);
            return true;

         case ss_Ctrl + 's':
            setSelectedWidget();
            setSortOrder(0);
            return true;

         case ss_Ctrl + 'p':
            setSelectedWidget();
            setSortOrder(1);
            return true;

         case ss_Ctrl + 'f':
            setSelectedWidget();
            setSortOrder(2);
            return true;

         case ss_Ctrl + 'c':
            setSelectedWidget();
            setSortOrder(3);
            return true;

         case ss_Ctrl + 'b':
            setSelectedWidget();
            setSortOrder(4);
            return true;

         case ss_Ctrl + 'x':
            setSelectedWidget();
            setSortOrder(5);
            return true;

         default:
            return UIWidget::handleEvent(evt, second_pass);
        }
    } else if (evt.type == ev_Mouse && isInside(GfxPoint(evt.mouse.x, evt.mouse.y))) {
        setSelectedWidget();
        if (evt.mouse.b != 0
            && evt.mouse.x < extent.x + label_width
            && evt.mouse.y >= extent.y + header_height
            && evt.mouse.y < extent.y + header_height + NUM_PLAYERS*line_height)
        {
            /* Click into left column */
            int pl = (evt.mouse.y - extent.y - header_height) / line_height;
            if (players[pl] != current_player)
                setCurrentPlayer(players[pl]);
        } else if (mouse_down
                   && evt.mouse.b == 0
                   && evt.mouse.y < extent.y + header_height)
        {
            /* Click into heading */
            int ord = (evt.mouse.x < extent.x + label_width
                       ? -1
                       : (evt.mouse.x - extent.x - label_width) / column_width);
            if (ord < int(columns.size()) && ord != sort_order)
                setSortOrder(ord);
        }
        mouse_down = (evt.mouse.b != 0);
        return true;
    } else {
        return UIWidget::handleEvent(evt, second_pass);
    }
}

void
WScoreTable::onStateChange(int bits, bool enabled)
{
    if ((bits & st_Selected) != 0 && !enabled)
        mouse_down = false;
    UIWidget::onStateChange(bits, enabled);
}

/** Get name of a mode.
    \param mode Mode number
    \param compare Record to compare. compare_record to show current mode, current_record to
    show mode for set */
string_t
WScoreTable::getModeName(Mode mode, uint32 compare)
{
    if (mode < countof(mode_names)) {
        return format(_(mode_names[mode]), file.getRecordByIndex(compare).getTurnNumber());
    } else {
        return "?";
    }
}

/** Update sort order. This populates the players[] array to match
    the current sort order. */
void
WScoreTable::updateSortOrder()
{
    const GStatRecord& record = file.getRecordByIndex(current_record);

    /* initialize */
    for (int i = 1; i <= NUM_PLAYERS; ++i)
        players[i-1] = i;

    /* sort, if requested */
    if (sort_order >= 0 && sort_order < int(columns.size())) {
        /* q&d bubblesort. Fast enough and avoids the need for a function object. */
        int p = 0;
        while (p < NUM_PLAYERS-1) {
            int32 a = columns[sort_order].get(record, players[p]);
            int32 b = columns[sort_order].get(record, players[p+1]);
            if (a == b) {
                a = columns[0].get(record, players[p]);
                b = columns[0].get(record, players[p+1]);
            }
            if (a < b) {
                std::swap(players[p], players[p+1]);
                if (p > 0)
                    --p;
            } else {
                ++p;
            }
        }
    }
}

/** Event handler: switch to next turn. */
void
WScoreTable::nextTurn()
{
    if (current_record+1 < file.getNumTurns())
        setCurrentRecord(current_record+1);
}

/** Event handler: switch to previous turn. */
void
WScoreTable::previousTurn()
{
    if (current_record > 0)
        setCurrentRecord(current_record-1);
}

/** Event handler: activate next player. */
void
WScoreTable::nextPlayer()
{
    if (current_player == 0) {
        /* No player selected, start with first line */
        setCurrentPlayer(players[0]);
    } else {
        /* Find next line */
        for (int i = 0; i < NUM_PLAYERS; ++i) {
            if (players[i] == current_player) {
                if (i+1 >= NUM_PLAYERS) {
                    /* After last line. Remove selection if possible, otherwise start at
                       first line. */
                    if (mode == Percentages)
                        setCurrentPlayer(players[0]);
                    else
                        setCurrentPlayer(0);
                } else {
                    /* Use next line */
                    setCurrentPlayer(players[i+1]);
                }
                break;
            }
        }
    }
}

/** Event handler: activate previous player. */
void
WScoreTable::previousPlayer()
{
    if (current_player == 0) {
        /* No player selected, start with last line */
        setCurrentPlayer(players[NUM_PLAYERS-1]);
    } else {
        /* Find previous line */
        for (int i = 0; i < NUM_PLAYERS; ++i) {
            if (players[i] == current_player) {
                if (i == 0) {
                    /* Before first line. Remove selection if possible, otherwise start at
                       last line. */
                    if (mode == Percentages)
                        setCurrentPlayer(players[NUM_PLAYERS-1]);
                    else
                        setCurrentPlayer(0);
                } else {
                    /* Use previous line */
                    setCurrentPlayer(players[i-1]);
                }
                break;
            }
        }
    }
}

/************************ WScoreTableConfigDisplay ***********************/

WScoreTableConfigDisplay::WScoreTableConfigDisplay(WScoreTable& table)
    : table(table)
{
    void (UIWidget::*borlandSucks)() = &UIWidget::drawWidget;
    table.sig_config_changed.add(static_cast<UIWidget*>(this), borlandSucks);
}      

void
WScoreTableConfigDisplay::getLayoutInfo(LayoutInfo& info)
{
    int max = 0;
    for (unsigned int i = 0; i < countof(mode_names); ++i) {
        int n = fonts[FONT_NORMAL]->getTextWidth(mode_names[i]);
        if (n > max)
            max = n;
    }

    info.pref_size = info.min_size = GfxPoint(max + 2*font_ems[FONT_NORMAL], font_heights[FONT_NORMAL]);
    info.flags = li_FlexibleH;
}

void
WScoreTableConfigDisplay::drawContent(GfxCanvas& can)
{
    UISkin& s = getSkin();
    s.startDraw();
    s.setSolid(true);
    s.setColor(UIColor::tc_Static);

    int x = extent.x;
    int y = extent.y + (extent.h - font_heights[FONT_NORMAL]) / 2;

    s.outTextF(can, FONT_NORMAL, x, y, extent.w, table.getModeName(table.getMode(), table.getComparedRecord()));

    can.update(extent);
}
