/**
  *  \file client/wtile.cc
  *
  *  (c) 2001-2009 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/wtile.h"
#include "ui/ui.h"
#include "ui/uiglobal.h"
#include "game/obj.h"
#include "ui/layout.h"

/********************************* WTile *********************************/

/** \class WTile
    \brief Widget displaying certain data (hooray.)

    Tiles can display the following stuff:
    - a constant heading
    - some fixed data
    - some variable data

    This class provides an interface intended to make display of such
    stuff easier. It automatically manages update events. In addition,
    a tile can request to have some associated controls (buttons).

    To avoid excess redraws, data is partitioned. Each tile knows on
    which events it wants to be updated (\c what parameter to
    constructor). Update events (see update(int)) include this
    parameter; the tile is redrawn if any of the bits in our \c what
    is set in the event, or if \c what is zero (meaning that we always
    want to be updated) or the event is zero (meaning that we don't
    know what was updated and thus update everything just-in-case).

    \todo I'm not 100% sure whether the redraw optimisations are still
    needed or whether we can simply update everything.

    (In PCC 1.x, the distinguishing property of a "Tile" was that it
    was freely moveable. In PCC 2, all widgets are moveable.) */

/** Standard constructor.
    \param size  size of this tile
    \param id    widget Id of this tile */
WTile::WTile(GfxPoint size, int id)
    : UIBaseComplexWidget(id),
      size(size),
      children(UINullLayout::instance),
      state(sContentOnly),
      layout(LeftJustified),
      lrmargin(0),
      children_move(false)
{
    addChildWidget(children, 0);
    computeCoordinates();
}

WTile::~WTile()
{ }

/** Draw the tile. Do not override in WTile descendants. */
void
WTile::drawContent(GfxCanvas& can)
{
    if (state != sContentOnly) {
        getSkin().startDraw();
        setColor(UIColor::tc_Heading);
        getSkin().setSolid(false);
        getSkin().outTextF(can, FONT_TITLE, extent.x, extent.y, extent.w, heading);
    }
    if (state != sHeadingOnly) {
        getSkin().startDraw();
        drawData(can);
    }
    can.update(getExtent());
}

/** Set tile layout. All child controls of the tile are arranged in a bounding box.
    Depending on the alignment, this box is pushed to the left or right side of the tile,
    and is either aligned with the content (and thus moves if the heading is switched on
    or off) or with the tile.
    \param layout   Alignment
    \param lrmargin Room to leave from left/right edge of tile, in pixels
    \param children_move true if children move with content, false if they are fixed relative to top of tile */
void
WTile::setLayout(Layout layout, int lrmargin, bool children_move)
{
    this->layout        = layout;
    this->lrmargin      = lrmargin;
    this->children_move = children_move;
}

/** Add control to tile.
    \param r Bounding rectangle relative to origin defined with setLayout(), see there.
    \param w Widget to add */
void
WTile::addControl(GfxRect r, UIBaseWidget& w)
{
    // coordinates are relative to children group
    r.moveBy(children.getExtent().getTopLeft());

    // add it
    w.setExtent(r);
    children.add(w);
}

void
WTile::onResize(GfxRect /*oldsize*/, bool /*by_layout*/)
{
    /* Compute bounding box of children. Relative point (0,0) is always included.
       This makes tiles work that look like this:
                        ,-----------------------------.
                        |         .------------------,|
         Bounding box ------------->             .--.||
                        |         |              |  <--- Widget
               Tile ------>       |              `--'||
                        |         `------------------'|
                        `-----------------------------'
       Those are created by having the leftmost widget not start at X=0. */
    GfxRect bbox(children.getExtent().getTopLeft(), GfxPoint(1, 1));
    for (UIBaseWidget* p = children.getFirstChild(); p != 0; p = p->getNextSibling())
        bbox.include(p->getExtent());

    /* Compute new position */
    GfxRect mysize = getExtent();
    if (layout == LeftJustified) {
        bbox.w += mysize.x - bbox.x;
        bbox.x =  mysize.x + lrmargin;
    } else {
        bbox.x = mysize.x + mysize.w - bbox.w - lrmargin;
    }
    bbox.y = mysize.y;
    if (children_move)
        bbox.y += y_origin;
    
    children.setExtent(bbox, true);
}

/** Set text color. Shorthand for setting the color in the skin. */
void
WTile::setColor(UISkin::Color c)
{
    getSkin().setColor(c);
}

/** Output text. Shorthand for UISkin::outText. In addition, this
    automatically does coordinate transformation. */
void
WTile::outText(GfxCanvas& can, int f, int x, int y, string_t text)
{
    getSkin().outText(can, f, x + extent.x, y + extent.y + y_origin, text);
}

/** Output text. Shorthand for UISkin::outTextF. In addition, this
    automatically does coordinate transformation. */
void
WTile::outTextF(GfxCanvas& can, int f, int x, int y, int lim, string_t text)
{
    getSkin().outTextF(can, f, x + extent.x, y + extent.y + y_origin, lim, text);
}

/** Draw background. Derived drawData() can use this to clear regions of
    the background. \c r is in relative coordinates. */
void
WTile::drawBackground(GfxCanvas& can, GfxRect r)
{
    r.moveBy(extent.x, extent.y + y_origin);
    r.intersect(extent);
    getSkin().drawBackground(can, r);
}

/** Get Layout Information. Do not override in WTile descendants. */
void
WTile::getLayoutInfo(LayoutInfo& info)
{
    GfxPoint size = this->size;
    if (state == sHeadingOnly)
        size.y = 0;
    size.y += y_origin;

    info.flags = 0;
    info.min_size = info.pref_size = size;
}

/** Get size of data area, with margins. This returns the dimensions
    of the data area you are expected to fill, in absolute coordinates.
    \param lmargin left margin (number of pixels to remove at the left)
    \param rmargin right margin */
GfxRect
WTile::getInnerSize(int lmargin, int rmargin)
{
    GfxRect rv = getExtent();
    rv.x += lmargin;
    rv.w -= lmargin+rmargin;
    rv.y += y_origin;
    rv.h -= y_origin;
    return rv;
}

/** Set display state.
    \todo Currently, this only works reliably when the tile is not
    yet in a group. It does not actually resize the tile. */
void
WTile::setDisplayState(State state)
{
    if (state == this->state)
        return;
    this->state = state;
    computeCoordinates();
    onResize(getExtent(), true);
    drawWidget();
    // FIXME: if widget is *really* already displayed, we probably
    // should tell our parent to resize us
}

/** Set heading of the tile.
    \return *this, for convenience */
WTile&
WTile::setHeading(string_t heading)
{
    this->heading = heading;
    drawWidget();
    return *this;
}

/** Compute coordinates, after a change to layout parameters. \internal */
void
WTile::computeCoordinates()
{
    if (state == sContentOnly) {
        y_origin = 0;
    } else {
        y_origin = font_heights[FONT_TITLE];
    }
}

/************************ WObjectTypeObserverTile ************************/

/*! \class WObjectTypeObserverTile
    \brief Tile observing an GObjectType's selected GObject

    The WObjectTypeObserverTile observes a GObjectType. Whenever a new object
    is selected on the GObjectType, or the selected object changes, the tile
    is redrawn. */

/** Constructor.
    \param size Size of tile
    \param type GObjectType to observe */
WObjectTypeObserverTile::WObjectTypeObserverTile(GfxPoint size, GObjectType& type)
    : WTile(size),
      type(type)
{
    conn_pos = type.sig_pos_changed.add(this, &WObjectTypeObserverTile::onPositionChanged);
    onPositionChanged();
}

WObjectTypeObserverTile::~WObjectTypeObserverTile()
{ }

/** Called when different object is selected on the GObjectType. */
void
WObjectTypeObserverTile::onPositionChanged()
{
    // reconnect object change signal, and update
    if (GObject* o = getObject()) {
        conn_obj = o->sig_changed.add(this, &WObjectTypeObserverTile::onObjectChanged);
    } else {
        conn_obj = 0;
    }
    onObjectChanged();
}

/** Called when the selected object changes. */
void
WObjectTypeObserverTile::onObjectChanged()
{
    drawWidget();
}

/** Get pointer to referred object. Note that this can be 0 when the GObjectType has
    no universe or the universe has no object of the required type. Ideally, this
    should be used as
    <code>if (GFoo* p = dynamic_cast<GFoo*>(getObject())) { ... }</code>;
    this way, it automatically copes with lacking objects or objects of different
    types. */
GObject*
WObjectTypeObserverTile::getObject()
{
    if (type.getUniverse())
        return &type.getObject(type.getCurrent());
    else
        return 0;
}
