/**
  *  \file game/config.cc
  *  \brief PConfig Management
  *
  *  This file contains functions to manage and parse PConfig. HConfig
  *  is a strict subset of PConfig (with the single exception of
  *  AllowVPAFeatures), so it is handled here, too.
  *
  *  (c) 2001-2009 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstdlib>
#include <cstring>
#include <cstdio>
#include <vector>
#include "cpluslib/assert.h"
#include "cpluslib/format.h"
#include "defs.h"
#include "game/config.h"
#include "game/gamedirs.h"
#include "game/host.h"
#include "game/struct.h"
#include "io/io.h"
#include "util/confpars.h"
#include "util/console.h"
#include "util/except.h"
#include "util/misc.h"
#include "util/nls.h"


ConfigCostOption::ConfigCostOption(Config& container, const char* name)
    : ConfigOption(container, name)
{ }

ConfigCostOption::~ConfigCostOption()
{ }

void
ConfigCostOption::set(string_t value)
{
    int count = 0;
    GCost last_value;
    string_t::size_type n;
    while (count < NUM_PLAYERS-1 && (n = value.find(',')) != string_t::npos) {
        last_value = GCost::fromString(value.substr(0, n));
        value.erase(0, n+1);
        data[count++] = last_value;
    }
    last_value = GCost::fromString(value);
    data[count++] = last_value;
    while (count < NUM_PLAYERS)
        data[count++] = last_value;
}

string_t
ConfigCostOption::toString() const
{
    string_t result = data[0].toPHostString();
    for (int i = 1; i < NUM_PLAYERS; ++i) {
        result += ",";
        result += data[i].toPHostString();
    }
    return result;
}

void
ConfigCostOption::set(int index, const GCost& value)
{
    ASSERT(index > 0 && index <= NUM_PLAYERS);
    data[index-1] = value;
}

GCost
ConfigCostOption::operator()(int player) const
{
    if (player > 0 && player <= NUM_PLAYERS)
        return data[player-1];
    else
        return data[NUM_PLAYERS-1];
}

static ValueIntParser parse_int;
static ValueBoolParser parse_bool, &parse_tristate = parse_bool, &parse_tristatex = parse_bool;

GGameConfig config;

/** Construct blank config object. */
GGameConfig::GGameConfig()
    : ConfigLevel(*this, "ConfigLevel", parse_int),
      PlayerRace(*this, "PlayerRace", parse_int),
      RecycleRate(*this, "RecycleRate", parse_int),
      RandomMeteorRate(*this, "RandomMeteorRate", parse_int),
      AllowMinefields(*this, "AllowMinefields", parse_bool),
      AllowAlchemy(*this, "AllowAlchemy", parse_bool),
      DeleteOldMessages(*this, "DeleteOldMessages", parse_bool),
      DisablePasswords(*this, "DisablePasswords", parse_bool),
      GroundKillFactor(*this, "GroundKillFactor", parse_int),
      GroundDefenseFactor(*this, "GroundDefenseFactor", parse_int),
      FreeFighters(*this, "FreeFighters", parse_int),
      RaceMiningRate(*this, "RaceMiningRate", parse_int),
      AllowBuildFighters(*this, "AllowBuildFighters", parse_bool),
      RebelsBuildFighters(*this, "RebelsBuildFighters", parse_bool),
      ColoniesBuildFighters(*this, "ColoniesBuildFighters", parse_bool),
      RobotsBuildFighters(*this, "RobotsBuildFighters", parse_bool),
      CloakFailureRate(*this, "CloakFailureRate", parse_int),
      RobCloakedShips(*this, "RobCloakedShips", parse_bool),
      ScanRange(*this, "ScanRange", parse_int),
      SensorRange(*this, "SensorRange", parse_int),
      DarkSenseRange(*this, "DarkSenseRange", parse_int),
      AllowHiss(*this, "AllowHiss", parse_bool),
      AllowRebelGroundAttack(*this, "AllowRebelGroundAttack", parse_bool),
      AllowSuperRefit(*this, "AllowSuperRefit", parse_bool),
      AllowWebMines(*this, "AllowWebMines", parse_bool),
      CloakFuelBurn(*this, "CloakFuelBurn", parse_int),
      AllowNewNatives(*this, "AllowNewNatives", parse_bool),
      AllowPlanetAttacks(*this, "AllowPlanetAttacks", parse_bool),
      BorgAssimilationRate(*this, "BorgAssimilationRate", parse_int),
      MineDecayRate(*this, "MineDecayRate", parse_int),
      WebMineDecayRate(*this, "WebMineDecayRate", parse_int),
      MaximumMinefieldRadius(*this, "MaximumMinefieldRadius", parse_int),
      MaximumWebMinefieldRadius(*this, "MaximumWebMinefieldRadius", parse_int),
      TransuraniumDecayRate(*this, "TransuraniumDecayRate", parse_int),
      StructureDecayPerTurn(*this, "StructureDecayPerTurn", parse_int),
      StructureDecayOnUnowned(*this, "StructureDecayOnUnowned", parse_int),
      ClimateLimitsPopulation(*this, "ClimateLimitsPopulation", parse_bool),
      AllowEatingSupplies(*this, "AllowEatingSupplies", parse_bool),
      AllowNoFuelMovement(*this, "AllowNoFuelMovement", parse_bool),
      MineHitOdds(*this, "MineHitOdds", parse_int),
      WebMineHitOdds(*this, "WebMineHitOdds", parse_int),
      MineScanRange(*this, "MineScanRange", parse_int),
      AllowMinesDestroyMines(*this, "AllowMinesDestroyMines", parse_bool),
      AllowEngineShieldBonus(*this, "AllowEngineShieldBonus", parse_bool),
      EngineShieldBonusRate(*this, "EngineShieldBonusRate", parse_int),
      FighterSweepRate(*this, "FighterSweepRate", parse_int),
      AllowColoniesSweepWebs(*this, "AllowColoniesSweepWebs", parse_bool),
      MineSweepRate(*this, "MineSweepRate", parse_int),
      WebMineSweepRate(*this, "WebMineSweepRate", parse_int),
      HissEffectRate(*this, "HissEffectRate", parse_int),
      RobFailureOdds(*this, "RobFailureOdds", parse_int),
      PlanetsAttackRebels(*this, "PlanetsAttackRebels", parse_bool),
      PlanetsAttackKlingons(*this, "PlanetsAttackKlingons", parse_bool),
      MineSweepRange(*this, "MineSweepRange", parse_int),
      WebMineSweepRange(*this, "WebMineSweepRange", parse_int),
      AllowScienceMissions(*this, "AllowScienceMissions", parse_bool),
      MineHitOddsWhenCloakedX10(*this, "MineHitOddsWhenCloakedX10", parse_int),
      DamageLevelForCloakFail(*this, "DamageLevelForCloakFail", parse_int),
      AllowFedCombatBonus(*this, "AllowFedCombatBonus", parse_bool),
      MeteorShowerOdds(*this, "MeteorShowerOdds", parse_int),
      MeteorShowerOreRanges(*this, "MeteorShowerOreRanges", parse_int),
      LargeMeteorsImpacting(*this, "LargeMeteorsImpacting", parse_int),
      LargeMeteorOreRanges(*this, "LargeMeteorOreRanges", parse_int),
      AllowMeteorMessages(*this, "AllowMeteorMessages", parse_bool),
      AllowOneEngineTowing(*this, "AllowOneEngineTowing", parse_bool),
      AllowHyperWarps(*this, "AllowHyperWarps", parse_bool),
      ClimateDeathRate(*this, "ClimateDeathRate", parse_int),
      AllowGravityWells(*this, "AllowGravityWells", parse_bool),
      CrystalsPreferDeserts(*this, "CrystalsPreferDeserts", parse_bool),
      AllowMinesDestroyWebs(*this, "AllowMinesDestroyWebs", parse_bool),
      MaxPlanetaryIncome(*this, "MaxPlanetaryIncome", parse_int),
      IonStormActivity(*this, "IonStormActivity", parse_int),
      AllowChunneling(*this, "AllowChunneling", parse_bool),
      AllowDeluxeSuperSpy(*this, "AllowDeluxeSuperSpy", parse_bool),
      IonStormsHideMines(*this, "IonStormsHideMines", parse_bool),
      AllowGloryDevice(*this, "AllowGloryDevice", parse_bool),
      AllowAntiCloakShips(*this, "AllowAntiCloakShips", parse_bool),
      AllowGamblingShips(*this, "AllowGamblingShips", parse_bool),
      AllowCloakedShipsAttack(*this, "AllowCloakedShipsAttack", parse_bool),
      AllowShipCloning(*this, "AllowShipCloning", parse_bool),
      AllowBoardingParties(*this, "AllowBoardingParties", parse_bool),
      AllowImperialAssault(*this, "AllowImperialAssault", parse_bool),
      RamScoopFuelPerLY(*this, "RamScoopFuelPerLY", parse_int),
      AllowAdvancedRefinery(*this, "AllowAdvancedRefinery", parse_bool),
      AllowBioscanners(*this, "AllowBioscanners", parse_bool),
      HullTechNotSlowedByMines(*this, "HullTechNotSlowedByMines", parse_int),
      UseAccurateFuelModel(*this, "UseAccurateFuelModel", parse_bool),
      DefenseForUndetectable(*this, "DefenseForUndetectable", parse_int),
      DefenseToBlockBioscan(*this, "DefenseToBlockBioscan", parse_int),
      FactoriesForDetectable(*this, "FactoriesForDetectable", parse_int),
      MinesForDetectable(*this, "MinesForDetectable", parse_int),
      FighterSweepRange(*this, "FighterSweepRange", parse_int),
      MineHitDamageFor100KT(*this, "MineHitDamageFor100KT", parse_int),
      WebHitDamageFor100KT(*this, "WebHitDamageFor100KT", parse_int),
      AllowRegisteredFunctions(*this, "AllowRegisteredFunctions", parse_bool),
      GravityWellRange(*this, "GravityWellRange", parse_int),
      Language(*this, "Language"),
      AllowPlayerMessages(*this, "AllowPlayerMessages", parse_bool),
      ScoringMethod(*this, "ScoringMethod"),
      TowedShipsBreakFree(*this, "TowedShipsBreakFree", parse_bool),
      NativeClimateDeathRate(*this, "NativeClimateDeathRate", parse_int),
      AllowMoreThan50Targets(*this, "AllowMoreThan50Targets", parse_bool),
      CrystalSinTempBehavior(*this, "CrystalSinTempBehavior", parse_bool),
      RGANeedsBeams(*this, "RGANeedsBeams", parse_bool),
      AllowRGAOnUnowned(*this, "AllowRGAOnUnowned", parse_bool),
      CPEnableLanguage(*this, "CPEnableLanguage", parse_bool),
      CPEnableBigTargets(*this, "CPEnableBigTargets", parse_bool),
      CPEnableRaceName(*this, "CPEnableRaceName", parse_bool),
      CPEnableAllies(*this, "CPEnableAllies", parse_tristate),
      CPEnableMessage(*this, "CPEnableMessage", parse_bool),
      AllowAnonymousMessages(*this, "AllowAnonymousMessages", parse_bool),
      DelayAllianceCommands(*this, "DelayAllianceCommands", parse_bool),
      TerraformRate(*this, "TerraformRate", parse_int),
      MaxColTempSlope(*this, "MaxColTempSlope", parse_int),
      WebDrainFuelLoss(*this, "WebDrainFuelLoss", parse_int),
      WebHitFuelLoss(*this, "WebHitFuelLoss", parse_int),
      AllowWormholes(*this, "AllowWormholes", parse_bool),
      WrmDisplacement(*this, "WrmDisplacement", parse_int),
      WrmRandDisplacement(*this, "WrmRandDisplacement", parse_int),
      WrmStabilityAddX10(*this, "WrmStabilityAddX10", parse_int),
      WrmRandStability(*this, "WrmRandStability", parse_int),
      WrmMassAdd(*this, "WrmMassAdd", parse_int),
      WrmRandMass(*this, "WrmRandMass", parse_int),
      WrmVoluntaryTravel(*this, "WrmVoluntaryTravel", parse_bool),
      WrmTravelDistDivisor(*this, "WrmTravelDistDivisor", parse_int),
      WrmTravelWarpSpeed(*this, "WrmTravelWarpSpeed", parse_int),
      WrmTravelCloaked(*this, "WrmTravelCloaked", parse_bool),
      WrmEntryPowerX100(*this, "WrmEntryPowerX100", parse_int),
      CPEnableGive(*this, "CPEnableGive", parse_tristate),
      AllowTowCloakedShips(*this, "AllowTowCloakedShips", parse_bool),
      RobCloakedChance(*this, "RobCloakedChance", parse_int),
      UnitsPerTorpRate(*this, "UnitsPerTorpRate", parse_int),
      UnitsPerWebRate(*this, "UnitsPerWebRate", parse_int),
      AllowESBonusAgainstPlanets(*this, "AllowESBonusAgainstPlanets", parse_bool),
      ShipCloneCostRate(*this, "ShipCloneCostRate", parse_int),
      AllowHyperjumpGravWells(*this, "AllowHyperjumpGravWells", parse_bool),
      NativeCombatSurvivalRate(*this, "NativeCombatSurvivalRate", parse_int),
      AllowPrivateerTowCapture(*this, "AllowPrivateerTowCapture", parse_bool),
      AllowCrystalTowCapture(*this, "AllowCrystalTowCapture", parse_bool),
      RoundGravityWells(*this, "RoundGravityWells", parse_bool),
      CPEnableSend(*this, "CPEnableSend", parse_bool),
      CumulativePillaging(*this, "CumulativePillaging", parse_bool),
      AllowInterceptAttack(*this, "AllowInterceptAttack", parse_bool),
      RaceGrowthRate(*this, "RaceGrowthRate", parse_int),
      ProductionRate(*this, "ProductionRate", parse_int),
      MineOddsWarpBonusX100(*this, "MineOddsWarpBonusX100", parse_int),
      CloakMineOddsWarpBonusX100(*this, "CloakMineOddsWarpBonusX100", parse_int),
      WebMineOddsWarpBonusX100(*this, "WebMineOddsWarpBonusX100", parse_int),
      MineTravelSafeWarp(*this, "MineTravelSafeWarp", parse_int),
      CloakedMineTravelSafeWarp(*this, "CloakedMineTravelSafeWarp", parse_int),
      WebMineTravelSafeWarp(*this, "WebMineTravelSafeWarp", parse_int),
      AllowCloakFailMessages(*this, "AllowCloakFailMessages", parse_bool),
      TonsScoreCountsPlanets(*this, "TonsScoreCountsPlanets", parse_bool),
      AllowExtendedMissions(*this, "AllowExtendedMissions", parse_bool),
      ExtMissionsStartAt(*this, "ExtMissionsStartAt", parse_int),
      WormholeUFOsStartAt(*this, "WormholeUFOsStartAt", parse_int),
      MaxShipsHissing(*this, "MaxShipsHissing", parse_int),
      SpyDetectionChance(*this, "SpyDetectionChance", parse_int),
      MapTruehullByPlayerRace(*this, "MapTruehullByPlayerRace", parse_bool),
      AllowWraparoundMap(*this, "AllowWraparoundMap", parse_bool),
      WraparoundRectangle(*this, "WraparoundRectangle", parse_int),
      CPEnableRemote(*this, "CPEnableRemote", parse_tristate),
      AllowAlliedChunneling(*this, "AllowAlliedChunneling", parse_bool),
      ColonistTaxRate(*this, "ColonistTaxRate", parse_int),
      NativeTaxRate(*this, "NativeTaxRate", parse_int),
      AllowAlternativeTowing(*this, "AllowAlternativeTowing", parse_bool),
      AllowBeamUpClans(*this, "AllowBeamUpClans", parse_bool),
      AllowBeamUpMultiple(*this, "AllowBeamUpMultiple", parse_bool),
      DamageLevelForTerraformFail(*this, "DamageLevelForTerraformFail", parse_int),
      DamageLevelForAntiCloakFail(*this, "DamageLevelForAntiCloakFail", parse_int),
      DamageLevelForChunnelFail(*this, "DamageLevelForChunnelFail", parse_int),
      DamageLevelForHyperjumpFail(*this, "DamageLevelForHyperjumpFail", parse_int),
      MinimumChunnelDistance(*this, "MinimumChunnelDistance", parse_int),
      TowStrengthEngineScale(*this, "TowStrengthEngineScale", parse_int),
      TowStrengthDistanceScale(*this, "TowStrengthDistanceScale", parse_int),
      AllowPriorityBuild(*this, "AllowPriorityBuild", parse_bool),
      SBQBuildPALBoost(*this, "SBQBuildPALBoost", parse_int),
      SBQNewBuildPALBoost(*this, "SBQNewBuildPALBoost", parse_int),
      SBQPointsForAging(*this, "SBQPointsForAging", parse_int),
      SBQBuildChangePenalty(*this, "SBQBuildChangePenalty", parse_int),
      SBQBoostExpX100(*this, "SBQBoostExpX100", parse_int),
      BuildChangeRelativePenalty(*this, "BuildChangeRelativePenalty", parse_int),
      PALDecayPerTurn(*this, "PALDecayPerTurn", parse_int),
      PALPlayerRate(*this, "PALPlayerRate", parse_int),
      PALCombatAggressor(*this, "PALCombatAggressor", parse_int),
      PALAggressorPointsPer10KT(*this, "PALAggressorPointsPer10KT", parse_int),
      PALOpponentPointsPer10KT(*this, "PALOpponentPointsPer10KT", parse_int),
      PALAggressorKillPointsPer10KT(*this, "PALAggressorKillPointsPer10KT", parse_int),
      PALOpponentKillPointsPer10KT(*this, "PALOpponentKillPointsPer10KT", parse_int),
      PALCombatPlanetScaling(*this, "PALCombatPlanetScaling", parse_int),
      PALCombatBaseScaling(*this, "PALCombatBaseScaling", parse_int),
      PALShipCapturePer10Crew(*this, "PALShipCapturePer10Crew", parse_int),
      PALRecyclingPer10KT(*this, "PALRecyclingPer10KT", parse_int),
      PALBoardingPartyPer10Crew(*this, "PALBoardingPartyPer10Crew", parse_int),
      PALGroundAttackPer100Clans(*this, "PALGroundAttackPer100Clans", parse_int),
      PALGloryDevice(*this, "PALGloryDevice", parse_int),
      PALGloryDevicePer10KT(*this, "PALGloryDevicePer10KT", parse_int),
      PALGloryDamagePer10KT(*this, "PALGloryDamagePer10KT", parse_int),
      PALGloryKillPer10KT(*this, "PALGloryKillPer10KT", parse_int),
      PALImperialAssault(*this, "PALImperialAssault", parse_int),
      PALRGA(*this, "PALRGA", parse_int),
      PALPillage(*this, "PALPillage", parse_int),
      PALIncludesESB(*this, "PALIncludesESB", parse_bool),
      FilterPlayerMessages(*this, "FilterPlayerMessages", parse_bool),
      AlternativeAntiCloak(*this, "AlternativeAntiCloak", parse_bool),
      AntiCloakImmunity(*this, "AntiCloakImmunity", parse_bool),
      AllowMoreThan500Minefields(*this, "AllowMoreThan500Minefields", parse_bool),
      NumMinefields(*this, "NumMinefields", parse_int),
      PALShipMinekillPer10KT(*this, "PALShipMinekillPer10KT", parse_int),
      MaximumMinefieldsPerPlayer(*this, "MaximumMinefieldsPerPlayer", parse_int),
      MineIdNeedsPermission(*this, "MineIdNeedsPermission", parse_bool),
      BuildQueue(*this, "BuildQueue"),
      PBPCostPer100KT(*this, "PBPCostPer100KT", parse_int),
      PBPMinimumCost(*this, "PBPMinimumCost", parse_int),
      PBPCloneCostRate(*this, "PBPCloneCostRate", parse_int),
      AllowShipNames(*this, "AllowShipNames", parse_tristate),
      BuildPointReport(*this, "BuildPointReport", parse_tristate),
      AlternativeMinesDestroyMines(*this, "AlternativeMinesDestroyMines", parse_bool),
      NumShips(*this, "NumShips", parse_int),
      ExtendedSensorSweep(*this, "ExtendedSensorSweep", parse_bool),
      ColonistCombatSurvivalRate(*this, "ColonistCombatSurvivalRate", parse_int),
      NewNativesPerTurn(*this, "NewNativesPerTurn", parse_int),
      NewNativesPopulationRange(*this, "NewNativesPopulationRange", parse_int),
      NewNativesRaceRate(*this, "NewNativesRaceRate", parse_int),
      NewNativesGovernmentRate(*this, "NewNativesGovernmentRate", parse_int),
      PlayerSpecialMission(*this, "PlayerSpecialMission", parse_int),
      TowedShipsCooperate(*this, "TowedShipsCooperate", parse_tristate),
      WrmScanRange(*this, "WrmScanRange", parse_int),
      FuelUsagePerFightFor100KT(*this, "FuelUsagePerFightFor100KT", parse_int),
      FuelUsagePerTurnFor100KT(*this, "FuelUsagePerTurnFor100KT", parse_int),
      CPEnableEnemies(*this, "CPEnableEnemies", parse_tristate),
      CPEnableShow(*this, "CPEnableShow", parse_tristate),
      CPEnableRefit(*this, "CPEnableRefit", parse_tristate),
      AllowIncompatibleConfiguration(*this, "AllowIncompatibleConfiguration", parse_bool),
      FreeFighterCost(*this, "FreeFighterCost"),
      StarbaseCost(*this, "StarbaseCost"),
      BaseFighterCost(*this, "BaseFighterCost"),
      ShipFighterCost(*this, "ShipFighterCost"),
      MaximumFightersOnBase(*this, "MaximumFightersOnBase", parse_int),
      MaximumDefenseOnBase(*this, "MaximumDefenseOnBase", parse_int),
      NumExperienceLevels(*this, "NumExperienceLevels", parse_int),
      ExperienceLevels(*this, "ExperienceLevels", parse_int),
      ExperienceLevelNames(*this, "ExperienceLevelNames"),
      ExperienceLimit(*this, "ExperienceLimit", parse_int),
      EPRecrewScaling(*this, "EPRecrewScaling", parse_int),
      EPShipAging(*this, "EPShipAging", parse_int),
      EPPlanetAging(*this, "EPPlanetAging", parse_int),
      EPPlanetGovernment(*this, "EPPlanetGovernment", parse_int),
      EPShipMovement100LY(*this, "EPShipMovement100LY", parse_int),
      EPShipHyperjump(*this, "EPShipHyperjump", parse_int),
      EPShipChunnel(*this, "EPShipChunnel", parse_int),
      EPShipIonStorm100MEV(*this, "EPShipIonStorm100MEV", parse_int),
      EPCombatKillScaling(*this, "EPCombatKillScaling", parse_int),
      EPCombatDamageScaling(*this, "EPCombatDamageScaling", parse_int),
      EPShipAlchemy100KT(*this, "EPShipAlchemy100KT", parse_int),
      EPCombatBoostRate(*this, "EPCombatBoostRate", parse_int),
      EPCombatBoostLevel(*this, "EPCombatBoostLevel", parse_int),
      EPTrainingScale(*this, "EPTrainingScale", parse_int),
      ExactExperienceReports(*this, "ExactExperienceReports", parse_bool),
      EModBayRechargeRate(*this, "EModBayRechargeRate", parse_int),
      EModBayRechargeBonus(*this, "EModBayRechargeBonus", parse_int),
      EModBeamRechargeRate(*this, "EModBeamRechargeRate", parse_int),
      EModBeamRechargeBonus(*this, "EModBeamRechargeBonus", parse_int),
      EModTubeRechargeRate(*this, "EModTubeRechargeRate", parse_int),
      EModBeamHitFighterCharge(*this, "EModBeamHitFighterCharge", parse_int),
      EModTorpHitOdds(*this, "EModTorpHitOdds", parse_int),
      EModBeamHitOdds(*this, "EModBeamHitOdds", parse_int),
      EModBeamHitBonus(*this, "EModBeamHitBonus", parse_int),
      EModStrikesPerFighter(*this, "EModStrikesPerFighter", parse_int),
      EModFighterBeamExplosive(*this, "EModFighterBeamExplosive", parse_int),
      EModFighterBeamKill(*this, "EModFighterBeamKill", parse_int),
      EModFighterMovementSpeed(*this, "EModFighterMovementSpeed", parse_int),
      EModMaxFightersLaunched(*this, "EModMaxFightersLaunched", parse_int),
      EModTorpHitBonus(*this, "EModTorpHitBonus", parse_int),
      EModTubeRechargeBonus(*this, "EModTubeRechargeBonus", parse_int),
      EModExtraFighterBays(*this, "EModExtraFighterBays", parse_int),
      EModEngineShieldBonusRate(*this, "EModEngineShieldBonusRate", parse_int),
      EModShieldDamageScaling(*this, "EModShieldDamageScaling", parse_int),
      EModShieldKillScaling(*this, "EModShieldKillScaling", parse_int),
      EModHullDamageScaling(*this, "EModHullDamageScaling", parse_int),
      EModCrewKillScaling(*this, "EModCrewKillScaling", parse_int),
      EModPlanetaryTorpsPerTube(*this, "EModPlanetaryTorpsPerTube", parse_int),
      EModMineHitOddsBonus(*this, "EModMineHitOddsBonus", parse_int),
      GameName(*this, "GameName"),
      BayRechargeRate(*this, "BayRechargeRate", parse_int),
      BayRechargeBonus(*this, "BayRechargeBonus", parse_int),
      BeamRechargeRate(*this, "BeamRechargeRate", parse_int),
      BeamRechargeBonus(*this, "BeamRechargeBonus", parse_int),
      TubeRechargeRate(*this, "TubeRechargeRate", parse_int),
      BeamHitFighterCharge(*this, "BeamHitFighterCharge", parse_int),
      BeamHitShipCharge(*this, "BeamHitShipCharge", parse_int),
      TorpFiringRange(*this, "TorpFiringRange", parse_int),
      BeamFiringRange(*this, "BeamFiringRange", parse_int),
      TorpHitOdds(*this, "TorpHitOdds", parse_int),
      BeamHitOdds(*this, "BeamHitOdds", parse_int),
      BeamHitBonus(*this, "BeamHitBonus", parse_int),
      StrikesPerFighter(*this, "StrikesPerFighter", parse_int),
      FighterKillOdds(*this, "FighterKillOdds", parse_int),
      FighterBeamExplosive(*this, "FighterBeamExplosive", parse_int),
      FighterBeamKill(*this, "FighterBeamKill", parse_int),
      ShipMovementSpeed(*this, "ShipMovementSpeed", parse_int),
      FighterMovementSpeed(*this, "FighterMovementSpeed", parse_int),
      BayLaunchInterval(*this, "BayLaunchInterval", parse_int),
      MaxFightersLaunched(*this, "MaxFightersLaunched", parse_int),
      AllowAlternativeCombat(*this, "AllowAlternativeCombat", parse_bool),
      StandoffDistance(*this, "StandoffDistance", parse_int),
      PlanetsHaveTubes(*this, "PlanetsHaveTubes", parse_bool),
      PlanetaryTorpsPerTube(*this, "PlanetaryTorpsPerTube", parse_int),
      FireOnAttackFighters(*this, "FireOnAttackFighters", parse_bool),
      TorpHitBonus(*this, "TorpHitBonus", parse_int),
      TubeRechargeBonus(*this, "TubeRechargeBonus", parse_int),
      ShieldDamageScaling(*this, "ShieldDamageScaling", parse_int),
      HullDamageScaling(*this, "HullDamageScaling", parse_int),
      CrewKillScaling(*this, "CrewKillScaling", parse_int),
      ShieldKillScaling(*this, "ShieldKillScaling", parse_int),
      ExtraFighterBays(*this, "ExtraFighterBays", parse_int),
      BeamHitFighterRange(*this, "BeamHitFighterRange", parse_int),
      FighterFiringRange(*this, "FighterFiringRange", parse_int),
      AllowVPAFeatures(*this, "AllowVPAFeatures", parse_int),
      CPEnableRumor(*this, "CPEnableRumor", AllowAnonymousMessages),
      RaceTaxRate(*this, "RaceTaxRate", ColonistTaxRate),
      CPNumMinefields(*this, "CPNumMinefields", NumMinefields),
      NativeClansRange(*this, "NativeClansRange", NewNativesPopulationRange),
      NativeTypeFrequencies(*this, "NativeTypeFrequencies", NewNativesRaceRate),
      NativeGovFrequencies(*this, "NativeGovFrequencies", NewNativesGovernmentRate)
{ }

GGameConfig::~GGameConfig()
{ }

/** Assign default values to all options. These defaults are (mostly) the PHost default
    values. A HCONFIG.HST file is never incomplete, so we need not deal with partially
    populated Tim-Host configuration at this point. */
void
GGameConfig::assignDefaults()
{
    ConfigLevel.set(0);
    PlayerRace.set("1,2,3,4,5,6,7,8,9,10,11");
    RecycleRate.set(75);
    RandomMeteorRate.set(2);
    AllowMinefields.set(1);
    AllowAlchemy.set(1);
    DeleteOldMessages.set(1);
    DisablePasswords.set(0);
    GroundKillFactor.set("1,30,1,15,1,1,1,1,1,1,1");
    GroundDefenseFactor.set("1,10,1,5,1,1,1,1,1,1,1");
    FreeFighters.set("0,0,0,0,0,0,0,10,0,0,0");
    RaceMiningRate.set("70,200,100,100,100,100,100,100,100,100,100");
    AllowBuildFighters.set("No,No,No,No,No,No,No,No,Yes,Yes,Yes");
    RebelsBuildFighters.set(1);
    ColoniesBuildFighters.set(1);
    RobotsBuildFighters.set(1);
    CloakFailureRate.set(1);
    RobCloakedShips.set(0);
    ScanRange.set(300);
    SensorRange.set(200);
    DarkSenseRange.copyFrom(SensorRange);
    AllowHiss.set(1);
    AllowRebelGroundAttack.set(1);
    AllowSuperRefit.set(1);
    AllowWebMines.copyFrom(AllowMinefields);
    CloakFuelBurn.set(5);
    AllowNewNatives.set(1);
    AllowPlanetAttacks.set(1);
    BorgAssimilationRate.set(100);
    MineDecayRate.set(5);
    WebMineDecayRate.copyFrom(MineDecayRate);
    MaximumMinefieldRadius.set(150);
    MaximumWebMinefieldRadius.copyFrom(MaximumMinefieldRadius);
    TransuraniumDecayRate.set(5);
    StructureDecayPerTurn.set(1);
    StructureDecayOnUnowned.set(1);
    ClimateLimitsPopulation.set(1);
    AllowEatingSupplies.set(0);
    AllowNoFuelMovement.set(1);
    MineHitOdds.set(1);
    WebMineHitOdds.set(5);
    MineScanRange.copyFrom(SensorRange);
    AllowMinesDestroyMines.set(1);
    AllowEngineShieldBonus.set(0);
    EngineShieldBonusRate.set(0);
    FighterSweepRate.set("0,0,0,0,0,0,0,0,0,0,20");
    AllowColoniesSweepWebs.set(0);
    MineSweepRate.set(4);
    WebMineSweepRate.set(3);
    HissEffectRate.set(5);
    RobFailureOdds.set(1);
    PlanetsAttackRebels.set(0);
    PlanetsAttackKlingons.set(0);
    MineSweepRange.set(10);
    WebMineSweepRange.set(5);
    AllowScienceMissions.set(1);
    MineHitOddsWhenCloakedX10.set(5);
    DamageLevelForCloakFail.set(1);
    AllowFedCombatBonus.set(1);
    MeteorShowerOdds.set(0);
    MeteorShowerOreRanges.set("10,10,10,10,200,200,200,200");
    LargeMeteorsImpacting.set(0);
    LargeMeteorOreRanges.set("100,100,100,100,10000,9000,9000,7000");
    AllowMeteorMessages.set(1);
    AllowOneEngineTowing.set(0);
    AllowHyperWarps.set(1);
    ClimateDeathRate.set(10);
    AllowGravityWells.set(1);
    CrystalsPreferDeserts.set(1);
    AllowMinesDestroyWebs.set(0);
    MaxPlanetaryIncome.set(5000);
    IonStormActivity.set(0);
    AllowChunneling.set(1);
    AllowDeluxeSuperSpy.set(1);
    IonStormsHideMines.set(1);
    AllowGloryDevice.set(1);
    AllowAntiCloakShips.set(1);
    AllowGamblingShips.set(1);
    AllowCloakedShipsAttack.set(1);
    AllowShipCloning.set(1);
    AllowBoardingParties.set(1);
    AllowImperialAssault.set(1);
    RamScoopFuelPerLY.set(2);
    AllowAdvancedRefinery.set(1);
    AllowBioscanners.set(1);
    HullTechNotSlowedByMines.set(7);
    UseAccurateFuelModel.set(0);
    DefenseForUndetectable.set(15);
    DefenseToBlockBioscan.set(20);
    FactoriesForDetectable.set(15);
    MinesForDetectable.set(20);
    FighterSweepRange.set("0,0,0,0,0,0,0,0,0,0,100");
    MineHitDamageFor100KT.set(100);
    WebHitDamageFor100KT.set(10);
    AllowRegisteredFunctions.set(1);
    GravityWellRange.set(3);
    Language.set("English");
    AllowPlayerMessages.set(1);
    ScoringMethod.set("Compatible");
    TowedShipsBreakFree.set(0);
    NativeClimateDeathRate.set(0);
    AllowMoreThan50Targets.set(0);
    CrystalSinTempBehavior.set(0);
    RGANeedsBeams.set(0);
    AllowRGAOnUnowned.set(0);
    CPEnableLanguage.set(1);
    CPEnableBigTargets.set(1);
    CPEnableRaceName.set(1);
    CPEnableAllies.set(1);
    CPEnableMessage.set(1);
    AllowAnonymousMessages.copyFrom(AllowPlayerMessages);
    DelayAllianceCommands.set(0);
    TerraformRate.set(1);
    MaxColTempSlope.set(1000);
    WebDrainFuelLoss.set(25);
    WebHitFuelLoss.set(50);
    AllowWormholes.set(1);
    WrmDisplacement.set(1);
    WrmRandDisplacement.set(0);
    WrmStabilityAddX10.set(0);
    WrmRandStability.set(0);
    WrmMassAdd.set(0);
    WrmRandMass.set(0);
    WrmVoluntaryTravel.set(1);
    WrmTravelDistDivisor.set(100);
    WrmTravelWarpSpeed.set(9);
    WrmTravelCloaked.set(0);
    WrmEntryPowerX100.set(25);
    CPEnableGive.set(1);
    AllowTowCloakedShips.set(0);
    RobCloakedChance.set(0);
    UnitsPerTorpRate.set("100,100,100,100,100,100,100,100,400,100,100");
    UnitsPerWebRate.copyFrom(UnitsPerTorpRate);
    AllowESBonusAgainstPlanets.set(1);
    ShipCloneCostRate.set("200,200,200,200,32767,200,32767,200,200,200,200");
    AllowHyperjumpGravWells.copyFrom(AllowGravityWells);
    NativeCombatSurvivalRate.set(75);
    AllowPrivateerTowCapture.copyFrom(AllowBoardingParties);
    AllowCrystalTowCapture.copyFrom(AllowBoardingParties);
    RoundGravityWells.set(0);
    CPEnableSend.set(1);
    CumulativePillaging.set(1);
    AllowInterceptAttack.set(1);
    RaceGrowthRate.set(100);
    ProductionRate.set(100);
    MineOddsWarpBonusX100.set(0);
    CloakMineOddsWarpBonusX100.set(0);
    WebMineOddsWarpBonusX100.set(0);
    MineTravelSafeWarp.set(0);
    CloakedMineTravelSafeWarp.copyFrom(MineTravelSafeWarp);
    WebMineTravelSafeWarp.copyFrom(MineTravelSafeWarp);
    AllowCloakFailMessages.set(1);
    TonsScoreCountsPlanets.set(0);
    AllowExtendedMissions.set(1);
    ExtMissionsStartAt.set(20);
    WormholeUFOsStartAt.set(51);
    MaxShipsHissing.set(500);
    SpyDetectionChance.set(20);
    MapTruehullByPlayerRace.set(0);
    AllowWraparoundMap.set(0);
    WraparoundRectangle.set("1000,1000,3000,3000");
    CPEnableRemote.set(1);
    AllowAlliedChunneling.set(1);
    ColonistTaxRate.set("200,100,100,100,100,100,100,100,100,100,100");
    NativeTaxRate.copyFrom(ColonistTaxRate);
    AllowAlternativeTowing.set(0);
    AllowBeamUpClans.set(1);
    AllowBeamUpMultiple.set(1);
    DamageLevelForTerraformFail.set(100);
    DamageLevelForAntiCloakFail.set(20);
    DamageLevelForChunnelFail.set(100);
    DamageLevelForHyperjumpFail.set(100);
    MinimumChunnelDistance.set(100);
    TowStrengthEngineScale.set(1);
    TowStrengthDistanceScale.set(19);
    AllowPriorityBuild.set(1);
    SBQBuildPALBoost.set(1);
    SBQNewBuildPALBoost.set(1);
    SBQPointsForAging.set(1200);
    SBQBuildChangePenalty.set(2147483647);
    SBQBoostExpX100.set(0);
    BuildChangeRelativePenalty.set(100);
    PALDecayPerTurn.set(20);
    PALPlayerRate.set(100);
    PALCombatAggressor.set(0);
    PALAggressorPointsPer10KT.set(2);
    PALOpponentPointsPer10KT.set(2);
    PALAggressorKillPointsPer10KT.set(10);
    PALOpponentKillPointsPer10KT.set(10);
    PALCombatPlanetScaling.set(50);
    PALCombatBaseScaling.set(80);
    PALShipCapturePer10Crew.set(5);
    PALRecyclingPer10KT.set(4);
    PALBoardingPartyPer10Crew.set(3);
    PALGroundAttackPer100Clans.set(100);
    PALGloryDevice.set(100);
    PALGloryDevicePer10KT.set(0);
    PALGloryDamagePer10KT.set(2);
    PALGloryKillPer10KT.set(0);
    PALImperialAssault.set(100);
    PALRGA.set(10);
    PALPillage.set(10);
    PALIncludesESB.set(1);
    FilterPlayerMessages.set(0);
    AlternativeAntiCloak.set(0);
    AntiCloakImmunity.set("Yes,Yes,Yes,No,No,No,No,No,No,No,No");
    AllowMoreThan500Minefields.set(0);
    NumMinefields.set(500);
    PALShipMinekillPer10KT.set(0);
    MaximumMinefieldsPerPlayer.set(10000);
    MineIdNeedsPermission.set(0);
    BuildQueue.set("PAL");
    PBPCostPer100KT.set(200);
    PBPMinimumCost.set(400);
    PBPCloneCostRate.set(200);
    AllowShipNames.set(1);
    BuildPointReport.set(2 /* Allies */);
    AlternativeMinesDestroyMines.set(0);
    NumShips.set(500);
    ExtendedSensorSweep.set(1);
    ColonistCombatSurvivalRate.set(0);
    NewNativesPerTurn.set(1);
    NewNativesPopulationRange.set("2500,5000");
    NewNativesRaceRate.set(1);
    NewNativesGovernmentRate.set(1);
    PlayerSpecialMission.copyFrom(PlayerRace);
    TowedShipsCooperate.set(1);
    WrmScanRange.set(100);
    FuelUsagePerFightFor100KT.set(0);
    FuelUsagePerTurnFor100KT.set(0);
    CPEnableEnemies.set(1);
    CPEnableShow.set(1);
    CPEnableRefit.set(1);
    AllowIncompatibleConfiguration.set(0);
    FreeFighterCost.set("T3 M2");
    StarbaseCost.set("T402 D120 M340 $900");
    BaseFighterCost.set("T3 M2 $100");
    ShipFighterCost.set("T3 M2 S5");
    MaximumFightersOnBase.set(60);
    MaximumDefenseOnBase.set(200);
    NumExperienceLevels.set(0);
    ExperienceLevels.set("750,1500,3000,6000");
    ExperienceLevelNames.set("Recruit,Soldier,Experienced,Elite,Ultra Elite");
    ExperienceLimit.set(1000000);
    EPRecrewScaling.set(30);
    EPShipAging.set(15);
    EPPlanetAging.set(25);
    EPPlanetGovernment.set(0);
    EPShipMovement100LY.set(40);
    EPShipHyperjump.set(30);
    EPShipChunnel.set(30);
    EPShipIonStorm100MEV.set(80);
    EPCombatKillScaling.set(800);
    EPCombatDamageScaling.set(200);
    EPShipAlchemy100KT.set(5);
    EPCombatBoostRate.set(100);
    EPCombatBoostLevel.set(0);
    EPTrainingScale.set(70);
    ExactExperienceReports.set(0);
    EModBayRechargeRate.set("1,2,3,4");
    EModBayRechargeBonus.set("0,0,0,0");
    EModBeamRechargeRate.set("0,0,0,0");
    EModBeamRechargeBonus.set("0,0,0,0");
    EModTubeRechargeRate.set("1,2,3,8");
    EModBeamHitFighterCharge.set("0,0,0,0");
    EModTorpHitOdds.set("9,18,27,35");
    EModBeamHitOdds.set("0,0,0,0");
    EModBeamHitBonus.set("0,0,0,0");
    EModStrikesPerFighter.set("1,2,3,4");
    EModFighterBeamExplosive.set("0,0,0,0");
    EModFighterBeamKill.set("0,0,0,0");
    EModFighterMovementSpeed.set("0,0,0,0");
    EModMaxFightersLaunched.set("0,0,0,0");
    EModTorpHitBonus.set("0,0,0,0");
    EModTubeRechargeBonus.set("0,0,0,0");
    EModExtraFighterBays.set("0,0,0,0");
    EModEngineShieldBonusRate.set("0,0,0,0");
    EModShieldDamageScaling.set("0,0,0,0");
    EModShieldKillScaling.set("0,0,0,0");
    EModHullDamageScaling.set("0,0,0,0");
    EModCrewKillScaling.set("-5,-10,-15,-20");
    EModPlanetaryTorpsPerTube.set("0,0,0,0");
    EModMineHitOddsBonus.set("5,10,15,20");
    GameName.set("Nameless Game");
    BayRechargeRate.set(52);
    BayRechargeBonus.set(2);
    BeamRechargeRate.set(8);
    BeamRechargeBonus.set(0);
    TubeRechargeRate.set(45);
    BeamHitFighterCharge.set(500);
    BeamHitShipCharge.set(600);
    TorpFiringRange.set(30000);
    BeamFiringRange.set(20000);
    TorpHitOdds.set(65);
    BeamHitOdds.set(100);
    BeamHitBonus.set(0);
    StrikesPerFighter.set(7);
    FighterKillOdds.set(20);
    FighterBeamExplosive.set(2);
    FighterBeamKill.set(2);
    ShipMovementSpeed.set(75);
    FighterMovementSpeed.set(235);
    BayLaunchInterval.set(3);
    MaxFightersLaunched.set(19);
    AllowAlternativeCombat.set(0);
    StandoffDistance.set(3000);
    PlanetsHaveTubes.set(0);
    PlanetaryTorpsPerTube.set(3);
    FireOnAttackFighters.set(0);
    TorpHitBonus.set(0);
    TubeRechargeBonus.set(0);
    ShieldDamageScaling.set(80);
    HullDamageScaling.set(80);
    CrewKillScaling.set(80);
    ShieldKillScaling.set(0);
    ExtraFighterBays.set("3,0,0,0,0,0,0,0,0,0,0");
    BeamHitFighterRange.set(100000);
    FighterFiringRange.set(3000);
    AllowVPAFeatures.set(1);

    markAllOptionsUnset();
}

/** Assign dependant options. If an option has the default "same as other option",
    and has not been explicitly set in the config file, copy its value from the other
    option. */
void
GGameConfig::assignDependantOptions()
{
    if (!DarkSenseRange.wasSet())
        DarkSenseRange.copyFrom(SensorRange);
    if (!AllowWebMines.wasSet())
        AllowWebMines.copyFrom(AllowMinefields);
    if (!WebMineDecayRate.wasSet())
        WebMineDecayRate.copyFrom(MineDecayRate);
    if (!MaximumWebMinefieldRadius.wasSet())
        MaximumWebMinefieldRadius.copyFrom(MaximumMinefieldRadius);
    if (!MineScanRange.wasSet())
        MineScanRange.copyFrom(SensorRange);
    if (!AllowAnonymousMessages.wasSet())
        AllowAnonymousMessages.copyFrom(AllowPlayerMessages);
    if (!UnitsPerWebRate.wasSet())
        UnitsPerWebRate.copyFrom(UnitsPerTorpRate);
    if (!AllowHyperjumpGravWells.wasSet())
        AllowHyperjumpGravWells.copyFrom(AllowGravityWells);
    if (!AllowPrivateerTowCapture.wasSet())
        AllowPrivateerTowCapture.copyFrom(AllowBoardingParties);
    if (!AllowCrystalTowCapture.wasSet())
        AllowCrystalTowCapture.copyFrom(AllowBoardingParties);
    if (!CloakedMineTravelSafeWarp.wasSet())
        CloakedMineTravelSafeWarp.copyFrom(MineTravelSafeWarp);
    if (!WebMineTravelSafeWarp.wasSet())
        WebMineTravelSafeWarp.copyFrom(MineTravelSafeWarp);
    if (!NativeTaxRate.wasSet())
        NativeTaxRate.copyFrom(ColonistTaxRate);
    if (!PlayerSpecialMission.wasSet())
        PlayerSpecialMission.copyFrom(PlayerRace);
}

/** Import 11 WORDs from HCONFIG image (per-player settings). */
static void
importArray16(ConfigStdOption& opt, const char data[], std::size_t index)
{
    for (int i = 1; i <= 11; ++i) {
        opt.set(i, getInt16(data, index));
        index += 2;
    }
}

/** Import 8 DWORDs from HCONFIG image (meteor settings). */
static void
importArray32(ConfigIntArrayOption<8>& opt, const char data[], std::size_t index)
{
    for (int i = 1; i <= 8; ++i) {
        opt.set(i, getInt32(data, index));
        index += 4;
    }
}

/** Assign all options from a HCONFIG.HST image.
    \param data [in] The image
    \param size [in] Size of image */
void
GGameConfig::assignFromHConfigImage(const char data[], std::size_t size)
{
    /* Instead of checking each option's position, we only check known
       version boundaries. */
    if (size >= 10) {
        RecycleRate.set(getUint16(data, 0)); RecycleRate.markSet();
        RandomMeteorRate.set(getUint16(data, 2)); RandomMeteorRate.markSet();
        AllowMinefields.set(getUint16(data, 4)); AllowMinefields.markSet();
        AllowAlchemy.set(getUint16(data, 6)); AllowAlchemy.markSet();
        DeleteOldMessages.set(getUint16(data, 8)); DeleteOldMessages.markSet();
    }
    if (size >= 186) {
        DisablePasswords.set(getUint16(data, 10)); DisablePasswords.markSet();
        importArray16(GroundKillFactor, data, 14); GroundKillFactor.markSet();
        importArray16(GroundDefenseFactor, data, 38); GroundDefenseFactor.markSet();
        importArray16(FreeFighters, data, 62); FreeFighters.markSet();
        importArray16(RaceMiningRate, data, 86); RaceMiningRate.markSet();
        importArray16(ColonistTaxRate, data, 110); ColonistTaxRate.markSet();
        RebelsBuildFighters.set(getUint16(data, 132)); RebelsBuildFighters.markSet();
        ColoniesBuildFighters.set(getUint16(data, 134)); ColoniesBuildFighters.markSet();
        RobotsBuildFighters.set(getUint16(data, 136)); RobotsBuildFighters.markSet();
        CloakFailureRate.set(getUint16(data, 138)); CloakFailureRate.markSet();
        RobCloakedShips.set(getUint16(data, 140)); RobCloakedShips.markSet();
        ScanRange.set(getUint16(data, 142)); ScanRange.markSet();
        DarkSenseRange.set(getUint16(data, 144)); DarkSenseRange.markSet();
        AllowHiss.set(getUint16(data, 146)); AllowHiss.markSet();
        AllowRebelGroundAttack.set(getUint16(data, 148)); AllowRebelGroundAttack.markSet();
        AllowSuperRefit.set(getUint16(data, 150)); AllowSuperRefit.markSet();
        AllowWebMines.set(getUint16(data, 152)); AllowWebMines.markSet();
        CloakFuelBurn.set(getUint16(data, 154)); CloakFuelBurn.markSet();
        SensorRange.set(getUint16(data, 156)); SensorRange.markSet();
        AllowNewNatives.set(getUint16(data, 158)); AllowNewNatives.markSet();
        AllowPlanetAttacks.set(getUint16(data, 160)); AllowPlanetAttacks.markSet();
        BorgAssimilationRate.set(getUint16(data, 162)); BorgAssimilationRate.markSet();
        WebMineDecayRate.set(getUint16(data, 164)); WebMineDecayRate.markSet();
        MineDecayRate.set(getUint16(data, 166)); MineDecayRate.markSet();
        MaximumMinefieldRadius.set(getUint16(data, 168)); MaximumMinefieldRadius.markSet();
        TransuraniumDecayRate.set(getUint16(data, 170)); TransuraniumDecayRate.markSet();
        StructureDecayPerTurn.set(getUint16(data, 172)); StructureDecayPerTurn.markSet();
        AllowEatingSupplies.set(getUint16(data, 174)); AllowEatingSupplies.markSet();
        AllowNoFuelMovement.set(getUint16(data, 176)); AllowNoFuelMovement.markSet();
        MineHitOdds.set(getUint16(data, 178)); MineHitOdds.markSet();
        WebMineHitOdds.set(getUint16(data, 180)); WebMineHitOdds.markSet();
        MineScanRange.set(getUint16(data, 182)); MineScanRange.markSet();
        AllowMinesDestroyMines.set(getUint16(data, 184)); AllowMinesDestroyMines.markSet();
    }
    if (size >= 288) {
        AllowEngineShieldBonus.set(getUint16(data, 186)); AllowEngineShieldBonus.markSet();
        EngineShieldBonusRate.set(getUint16(data, 188)); EngineShieldBonusRate.markSet();
        AllowColoniesSweepWebs.set(getUint16(data, 192)); AllowColoniesSweepWebs.markSet();
        MineSweepRate.set(getUint16(data, 194)); MineSweepRate.markSet();
        WebMineSweepRate.set(getUint16(data, 196)); WebMineSweepRate.markSet();
        HissEffectRate.set(getUint16(data, 198)); HissEffectRate.markSet();
        RobFailureOdds.set(getUint16(data, 200)); RobFailureOdds.markSet();
        PlanetsAttackRebels.set(getUint16(data, 202)); PlanetsAttackRebels.markSet();
        PlanetsAttackKlingons.set(getUint16(data, 204)); PlanetsAttackKlingons.markSet();
        MineSweepRange.set(getUint16(data, 206)); MineSweepRange.markSet();
        WebMineSweepRange.set(getUint16(data, 208)); WebMineSweepRange.markSet();
        AllowScienceMissions.set(getUint16(data, 210)); AllowScienceMissions.markSet();
        MineHitOddsWhenCloakedX10.set(getUint16(data, 212)); MineHitOddsWhenCloakedX10.markSet();
        DamageLevelForCloakFail.set(getUint16(data, 214)); DamageLevelForCloakFail.markSet();
        AllowFedCombatBonus.set(getUint16(data, 216)); AllowFedCombatBonus.markSet();
        MeteorShowerOdds.set(getUint16(data, 218)); MeteorShowerOdds.markSet();
        importArray32(MeteorShowerOreRanges, data, 220); MeteorShowerOreRanges.markSet();
        LargeMeteorsImpacting.set(getUint16(data, 252)); LargeMeteorsImpacting.markSet();
        importArray32(LargeMeteorOreRanges, data, 254); LargeMeteorOreRanges.markSet();
        AllowMeteorMessages.set(getUint16(data, 286)); AllowMeteorMessages.markSet();
    }
    if (size >= 298) {
        AllowOneEngineTowing.set(getUint16(data, 288)); AllowOneEngineTowing.markSet();
        AllowHyperWarps.set(getUint16(data, 290)); AllowHyperWarps.markSet();
        ClimateDeathRate.set(getUint16(data, 292)); ClimateDeathRate.markSet();
        AllowGravityWells.set(getUint16(data, 294)); AllowGravityWells.markSet();
        CrystalsPreferDeserts.set(getUint16(data, 296)); CrystalsPreferDeserts.markSet();
    }
    if (size >= 302) {
        AllowMinesDestroyWebs.set(getUint16(data, 298)); AllowMinesDestroyWebs.markSet();
        ClimateLimitsPopulation.set(getUint16(data, 300)); ClimateLimitsPopulation.markSet();
    }
    if (size >= 328) {
        MaxPlanetaryIncome.set(getUint32(data, 302)); MaxPlanetaryIncome.markSet();
        IonStormActivity.set(getUint16(data, 306)); IonStormActivity.markSet();
        AllowChunneling.set(getUint16(data, 308)); AllowChunneling.markSet();
        AllowDeluxeSuperSpy.set(getUint16(data, 310)); AllowDeluxeSuperSpy.markSet();
        IonStormsHideMines.set(getUint16(data, 312)); IonStormsHideMines.markSet();
        AllowGloryDevice.set(getUint16(data, 314)); AllowGloryDevice.markSet();
        AllowAntiCloakShips.set(getUint16(data, 316)); AllowAntiCloakShips.markSet();
        AllowGamblingShips.set(getUint16(data, 318)); AllowGamblingShips.markSet();
        AllowCloakedShipsAttack.set(getUint16(data, 320)); AllowCloakedShipsAttack.markSet();
        AllowShipCloning.set(getUint16(data, 322)); AllowShipCloning.markSet();
        AllowBoardingParties.set(getUint16(data, 324)); AllowBoardingParties.markSet();
        AllowImperialAssault.set(getUint16(data, 326)); AllowImperialAssault.markSet();
    }
    if (size >= 336) {
        RamScoopFuelPerLY.set(getUint16(data, 328)); RamScoopFuelPerLY.markSet();
        AllowAdvancedRefinery.set(getUint16(data, 330)); AllowAdvancedRefinery.markSet();
        AllowBioscanners.set(getUint16(data, 332)); AllowBioscanners.markSet();
        HullTechNotSlowedByMines.set(getUint16(data, 334)); HullTechNotSlowedByMines.markSet();
    }
    if (size >= 340) {
        AllowVPAFeatures.set(getUint16(data, 338)); AllowVPAFeatures.markSet();
    }
}

/** Load PCONFIG.SRC.
    \param pconf pconfig.src file
    \param shiplist shiplist.txt file, may be null. */
void
loadPConfig(Ptr<Stream> pconf, Ptr<Stream> shiplist)
{
    // FIXME: remove Ptr<> here.
    ConfigFileParser parser(config, true);
    console.write(LOG_INFO, format(_("Reading configuration from %s..."), pconf->getName()));
    parser.configure("phost", true);
    parser.parseFile(*pconf);
    if (shiplist) {
        console.write(LOG_INFO, format(_("Reading configuration from %s..."), shiplist->getName()));
        parser.configure("phost", false);
        parser.parseFile(*shiplist);
    }
    config.assignDependantOptions();
    host.setHavePConfig();
}

/** Load HCONFIG.HST.
    \todo do host version guessing here? */
void
loadHConfig(Stream& hconf)
{
    Stream::fpos_t size = hconf.getSize();
    if (size < 0 || size > CONF_MAX_HCONFIG_SIZE)
        throw FileFormatException(hconf, _("File has invalid size"));

    ArrayPtr<char> data (size);
    hconf.readT(data.ptr(), size);

    console.write(LOG_INFO, format(_("Reading configuration from %s..."), hconf.getName()));
    config.assignFromHConfigImage(data.ptr(), size);
    config.assignDependantOptions();
    host.setHaveHConfig(size);
}

/** Load default config. Initializes the config variable with all
    default values. The defaults are the default values from PHost,
    which should also be the THost defaults. */
void
loadDefaultConfig()
{
    config.assignDefaults();
}

/** Initialize configuration. Loads all configuration files. */
void
initConfig()
{
    Ptr<Stream> ps;

    loadDefaultConfig();

    ps = spec_file_dir->openFileNT("pconfig.src", Stream::C_READ);
    if (ps.ptr()) {
        loadPConfig(ps, spec_file_dir->openFileNT("shiplist.txt", Stream::C_READ));
        return;
    }
    ps = spec_file_dir->openFileNT("hconfig.hst", Stream::C_READ);
    if (ps.ptr()) {
        loadHConfig(*ps);
        return;
    }
    console.write(LOG_WARN, _("No host configuration file found -- using defaults"));
}

/** Get name of an experience level.
    \param nr level number (0..NUM_EXPERIENCE_LEVELS) */
string_t
getExperienceLevelName(int nr)
{
    string_t s = config.ExperienceLevelNames();
    for (int i = 0; i < nr; ++i)
        strRemove(s, ",");
    s = strTrim(strFirst(s, ","));
    if (s.size())
        return s;
    else
        return format(_("Level %d"), nr);
}

/** Get set of all players of a particular race. */
GPlayerSet
getPlayersOfRace(int nr)
{
    return getPlayersWhere(config.PlayerRace, nr);
}

/** Get set of all players where an option is enabled.
    \param opt Boolean option */
GPlayerSet
getPlayersWhereEnabled(const ConfigStdOption& opt)
{
    GPlayerSet result;
    for (int i = 1; i <= NUM_PLAYERS; ++i)
        if (opt(i) != 0)
            result += i;
    return result;
}

/** Get set of all players where an option has a given value.
    \param opt Option
    \param value Value to check for */
GPlayerSet
getPlayersWhere(const ConfigStdOption& opt, int value)
{
    GPlayerSet result;
    for (int i = 1; i <= NUM_PLAYERS; ++i)
        if (opt(i) == value)
            result += i;
    return result;
}
