/**
  *  \file game/cost.cc
  *  \brief Cost/Resource Set
  *
  *  (c) 2001-2009 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/cost.h"
#include "cpluslib/format.h"

/** Constructor. Construct empty (zero) cost. */
GCost::GCost()
    : t(0), d(0), m(0), mc(0), sup(0)
{ }

/** Constructor. Construct from TCost structure, as stored in
    specification file. */
GCost::GCost(const TCost& cost)
    : t(cost.t), d(cost.d), m(cost.m), mc(cost.mc), sup(0)
{ }

/** Try to sell supplies to cover a money shortage. */
void
GCost::trySupplySale()
{
    if (mc < 0) {
        sup += mc;
        mc = 0;
        if (sup < 0) {
            // not enough supplies - leave it a money shortage
            mc = sup;
            sup = 0;
        }
    }
}

/** Add costs.
    \param other Value to add to this one.
    \pre other.isValid() */
GCost&
GCost::operator+=(const GCost& other)
{
    t += other.t;
    d += other.d;
    m += other.m;
    mc += other.mc;
    sup += other.sup;
    return *this;
}

/** Multiply in-place.
    \param n Factor to multiply this with. */
GCost&
GCost::operator*=(int n)
{
    t *= n;
    d *= n;
    m *= n;
    mc *= n;
    sup *= n;
    return *this;
}

/** Multiply.
    \param n Factor to multiply with */
GCost
GCost::operator*(int n) const
{
    GCost result(*this);
    result *= n;
    return result;
}

/** Compare for equality.
    \param other other value
    \return true iff these two costs are equal */
bool
GCost::operator==(const GCost& other) const
{
    return t == other.t
        && d == other.d
        && m == other.m
        && mc == other.mc
        && sup == other.sup;
}

/** Subtract. This automatically handles supply-sale.
    \param other Value to subtract from this one
    \pre other.isValid() */
GCost&
GCost::operator-=(const GCost& other)
{
    t -= other.t;
    d -= other.d;
    m -= other.m;
    mc -= other.mc;
    sup -= other.sup;
    trySupplySale();
    return *this;
}

/** Set component.
    \param type Component to set. Must be el_Tritanium, el_Duranium,
                el_Molybdenum, el_Money, el_Supplies.
    \param n Value to set component to */
void
GCost::set(GCargoType type, int32 n)
{
    switch (type) {
     case el_Tritanium:  t = n; break;
     case el_Duranium:   d = n; break;
     case el_Molybdenum: m = n; break;
     case el_Money:     mc = n; break;
     case el_Supplies: sup = n; break;
     default: ASSERT(0); // FIXME?
    }
}

/** Get component.
    \param type component to query
    \return amount of component (negative if missing); 0 if type is outside
    valid range */
int32
GCost::get(GCargoType type) const
{
    switch (type) {
     case el_Tritanium: return t;
     case el_Duranium: return d;
     case el_Molybdenum: return m;
     case el_Money: return mc;
     case el_Supplies: return sup;
     default: return 0;
    }
}

/** Add component.
    \param type component to add.
    \param n increment */
void
GCost::add(GCargoType type, int32 n)
{
    set(type, get(type) + n);
}

/** Remove component.
    \param type component to remove.
    \param n decrement */
void
GCost::remove(GCargoType type, int32 n)
{
    set(type, get(type) - n);
}

/** Check whether this resource amount is large enough to buy an item.
    \param other price of the item
    \return true iff this cost is equal or higher than \c other. Supply sale
    is handled. */
bool
GCost::isEnoughFor(const GCost& other) const
{
    GCost tmp(*this);
    tmp -= other;
    return tmp.isValid();
}

/** Check whether this cost is valid.
    \return true iff all components are >= 0 */
bool
GCost::isValid() const
{
    return t >= 0 && d >= 0 && m >= 0 && mc >= 0 && sup >= 0;
}

static void
addItem(string_t& result, char letter, int32 value)
{
    if (value != 0) {
        if (result.size())
            result += ' ';
        result += letter;
        result += format("%d", value);
    }
}

/** Convert to PHost-style string. */
string_t
GCost::toPHostString() const
{
    string_t result;
    addItem(result, 'T', t);
    addItem(result, 'D', d);
    addItem(result, 'M', m);
    addItem(result, '$', mc);
    addItem(result, 'S', sup);
    if (result.empty())
        result = "S0";
    return result;
}

/** Parse number from string.
    \param value [in] the string
    \param i [in/out] current parse position
    \return the number */
static int32
eatNumber(const string_t& value, string_t::size_type& i)
{
    int32 amount = 0;
    while (i < value.size() && value[i] >= '0' && value[i] <= '9') {
        amount = 10*amount + value[i] - '0';
        ++i;
    }
    return amount;
}

/** Convert character into cargo type.
    \return el_XXX on success, -1 on failure. */
static int
charToCargoType(char c)
{
    switch (c) {
     case 'T': case 't':
        return el_Tritanium;
     case 'D': case 'd':
        return el_Duranium;
     case 'M': case 'm':
        return el_Molybdenum;
     case '$':
        return el_Money;
     case 'S': case 's':
        return el_Supplies;
     default:
        return -1;
    }
}

/** Parse a string into a GCost structure. This is intended to be used
    by the script and config interface. */
GCost
GCost::fromString(const string_t value)
{
    GCost result;

    /* Like PCC 1.x, we accept cargospecs and PHost format:
          Cargospec:  123TDM
          PHost:      T123 D123 M123 */
    string_t::size_type i = 0;
    while (1) {
        /* skip whitespace */
        while (i < value.size() && (value[i] == ' ' || value[i] == '\t'))
            ++i;
        if (i >= value.size())
            break;

        if (value[i] >= '0' && value[i] <= '9') {
            /* Cargospec */
            int32 amount = eatNumber(value, i);
            int type;
            while (i < value.size() && (type = charToCargoType(value[i])) >= 0) {
                result.set(GCargoType(type), result.get(GCargoType(type)) + amount);
                ++i;
            }
        } else {
            /* Must be PHost format */
            int type = charToCargoType(value[i]);
            ++i;
            if (type < 0 || i >= value.size()) {
                /* at this point, we know that it has a wrong format. 
                   FIXME: what can we do? Should we report an error? 
                   PCC 1.x just exits, so we now do that as well. */
                break;
            }
            int32 amount = eatNumber(value, i);
            result.set(GCargoType(type), result.get(GCargoType(type)) + amount);
        }
    }

    return result;
}
