/**
  *  \file game/mine.cc
  *  \brief Minefield Implementation
  *
  *  (c) 2001-2009 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cmath>
#include "game/mine.h"
#include "game/univ.h"
#include "util/nls.h"
#include "game/config.h"
#include "game/formula.h"
#include "cpluslib/format.h"
#include "game/host.h"
#include "util/misc.h"

/** Create empty minefield with specified Id. */
GMinefield::GMinefield(int aid, int ac)
    : id(aid), x(0), y(0), owner(0),
      web(false), units(0), turn(0), last_action(ac),
      units_prev(0), turn_prev(0),
      radius_now(0), units_now(0)
{ }

/** Postprocessing after load. */
void
GMinefield::internalCheck(GUniverse& univ)
{
    int32 u = units;
    for (int i = turn; i < univ.getTurnNumber(); ++i)
        u = getUnitsAfterDecay(u);
    units_now = u;
    radius_now = getRadiusFromUnits(units_now);
}

/** Compute mine decay for one turn. */
int32
GMinefield::getUnitsAfterDecay(int32 orig_units)
{
    // FIXME
    int decay_rate = web ? config.WebMineDecayRate(playerIndex(owner)) : config.MineDecayRate(playerIndex(owner));
    if (host.isPHost())
        /* PHost formula */
        return orig_units * (100 - decay_rate) / 100;
    else
        /* THost formula (3.22.040). Actual formula is
             ERND(orig_units - orig_units*decay_rate/100)
           which should yield the same results. */
        return rdivaddl(orig_units * (100-decay_rate), 100, 0);
}

/** Compute minefield radius from unit number. */
int
GMinefield::getRadiusFromUnits(int32 units)
{
    return int(std::sqrt(double(units)));
}

/** Compute successful passage rate. This is the inverse of the "hit
    rate", for a given distance. Under THost, there is only "the" hit
    rate. Under PHost, various options are arrayized, and dynamic by
    experience and speed. This function returns the value for an
    inexperienced ship owned by the current player, at warp 9 (=the
    worst possible case).

    \param distance Distance to cover, in ly
    \param cloaked  Whether ship is cloaked
    \return Passage rate, [0.0, 1.0] */
double
GMinefield::getPassRate(double distance, bool cloaked)
{
    double rate = (isWeb()
                   ? config.WebMineHitOdds(getPlayerId()) * 0.01
                   : cloaked
                   ? config.MineHitOddsWhenCloakedX10(getPlayerId()) * 0.001
                   : config.MineHitOdds(getPlayerId()) * 0.01);
    if (rate <= 0)
        /* Hit rate below zero -> pass rate is 1.0 */
        return 1.0;
    else if (rate >= 1.0)
        /* Hit rate is one -> pass rate is 0.0 */
        return 0.0;
    else
        /* Normal computation */
        return pow(1.0 - rate, distance);
}

/** Set data. Use this to "assimilate" a minefield report into the
    database. */
void
GMinefield::setData(int ax, int ay, int aown, int32 aunits, int aturn)
{
    if (aturn != turn) {
        if (x == ax && y == ay) {
            turn_prev = turn;
            units_prev = units;
        } else {
            turn_prev = 0;
            units_prev = 0;
        }
    }
    x     = ax;
    y     = ay;
    owner = aown;
    units = aunits;
    turn  = aturn;
}

/** Set data. Use this to "assimilate" a minefield report into the
    database. This also sets the minefield type. */
void
GMinefield::setData(int ax, int ay, int aown, int32 aunits, int aturn, bool aweb)
{
    setData(ax, ay, aown, aunits, aturn);
    web = aweb;
}

string_t
GMinefield::getName() const
{
    if (web)
        return format(_("Web Mine Field #%d"), id);
    else
        return format(_("Mine Field #%d"), id);
}

string_t
GMinefield::getFullName() const
{
    return getName();
}

int
GMinefield::getId() const
{
    return id;
}

int
GMinefield::getOwner() const
{
    return owner;
}

bool
GMinefield::isExistant() const
{
    return units_now != 0;
}

GPoint
GMinefield::getPos() const
{
    return GPoint(x, y);
}

int
GMinefield::getRadius() const
{
    return radius_now;
}

/************************** GMinefieldContainer **************************/

GMinefieldContainer::GMinefieldContainer()
{ }

GMinefieldContainer::~GMinefieldContainer()
{
    clear();
}

/** Get minefield, creating it if necessary.
    \param id Id of minefield
    \param cause Reason why to create this minefield. */
GMinefield&
GMinefieldContainer::getCreate(int id, int cause)
{
    minemap_t::iterator i = mines.find(id);
    if (i == mines.end()) {
        i = safeInsert(mines, id, new GMinefield(id, cause));
    } else {
        if (cause > i->second->getAction())
            i->second->setAction(cause);
    }
    return *i->second;
}

/** Get minefield if it exists.
    \param id Id of minefield to search */
GMinefield*
GMinefieldContainer::get(int id) const
{
    minemap_t::const_iterator i = mines.find(id);
    if (i == mines.end())
        return 0;
    else
        return i->second;
}

/** Erase minefield by iterator.
    \param i Iterator to minefield to erase
    \return iterator to next minefield */
GMinefieldContainer::iterator
GMinefieldContainer::erase(iterator i)
{
    // return value
    iterator p = i;
    ++p;

    // erase
    minemap_t::iterator base = i.base().base();
    delete base->second;
    mines.erase(base);

    return p;
}

/** Erase minefield by Id.
    \param id Id of minefield to erase. */
void
GMinefieldContainer::erase(int id)
{
    minemap_t::iterator p = mines.find(id);
    if (p != mines.end()) {
        delete p->second;
        mines.erase(p);
    }
}

/** Forget all minefields. */
void
GMinefieldContainer::clear()
{
    for (minemap_t::iterator i = mines.begin(); i != mines.end(); ++i)
        delete i->second;
    minemap_t().swap(mines);
}

/***************************** GMinefieldType ****************************/

// FIXME: count is hardcoded to 10000 which is the PHost limit.
// We can do without this by deriving from GObjectType and implementing
// our own getNext()/getPrev().
GMinefieldType::GMinefieldType()
    : GIndexedObjectType(10000)
{ }

bool
GMinefieldType::isExistant(int index)
{
    return getUniverse() != 0
        && getUniverse()->getMinefields().get(index) != 0;
}

GMinefield&
GMinefieldType::getObject(int index)
{
    return *getUniverse()->getMinefields().get(index);
}
