/**
  *  \file game/objtypes.cc
  *  \brief Object types
  *
  *  (c) 2001-2009 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/objtypes.h"
#include "game/obj.h"
#include "cpluslib/assert.h"

/****************************** GObjectType ******************************/

/** \class GObjectType
    \brief Object Type

    A class derived from GObjectType defines a set of objects, and
    iteration through them. In addition, it contains a general
    interface for object selection. Control screens and various other
    features use these as a back-end.

    A GObjectType object also manages a "current position".

    A GObjectType object starts without being associated to a universe.
    It can be associated with one by calling setUniverse(). The only
    derived-class method that has to deal with a null universe is init().
    The other methods can assume that a universe is present. Note that
    this implies the user shouldn't call any interesting method while
    the GObjectType is not associated with a universe.

    \todo We currently do not interpret any meaning into object Ids other
    than that 0 means no object. Do we need to be more strict? I guess we
    don't, and can just let GPlanetType etc. guarantee the additional
    requirement that id == got.getObject(id).getId(). Not having this
    requirement simplifies implementing the list for the visual scanner,
    for example. */

/** Initialize.
    \param max  Maximum object Id.
    \param trn  Related universe object. */
GObjectType::GObjectType()
    : current(0), universe(0)
{ }

GObjectType::~GObjectType()
{ }

/** Set current object Id.
    \pre isExistant(index) */
void
GObjectType::setCurrent(int index)
{
    ASSERT(isExistant(index));
    if (current != index) {
        current = index;
        sig_pos_changed.raise(index);
    }
}

int
GObjectType::getCurrent() const
{
    return current;
}

/** Change universe. This can also be used to trigger re-scan of an existing universe. */
void
GObjectType::setUniverse(GUniverse* univ)
{
    this->universe = univ;
    this->init();
    if (univ == 0) {
        // Universe removed
        current = 0;
    } else {
        // Universe changed
        if (!isExistant(current))
            current = getNext(0, false);
    }
    sig_pos_changed.raise(current);
}

GUniverse*
GObjectType::getUniverse() const
{
    return universe;
}

/*************************** GIndexedObjectType **************************/

/** \class GIndexedObjectType
    \brief Indexed Object Type

    This refines GObjectType to represent a set of objects represented as a
    (sparse) object array. Objects are arranged in Id order. There is a finite
    set of Id numbers ranging from 1 to a given maximum. */

/** Constructor.
    \param max Highest possible Id number. */
GIndexedObjectType::GIndexedObjectType(int max)
    : count(0), max(max)
{ }

void
GIndexedObjectType::init()
{
    GUniverse* u = getUniverse();
    count = 0;
    if (u != 0) {
        for (int i = 1; i <= max; ++i) {
            if (isExistant(i)) {
                ++count;
            }
        }
    }
}

int
GIndexedObjectType::getNext(int index, bool mark)
{
    bool wrapped = false;
    while (1) {
        if (++index > max) {
            index = 0;
            if (wrapped)
                return 0;
            wrapped = 1;
        }
        if (isExistant(index) && (!mark || getObject(index).isMarked()))
            return index;
    }
}

int
GIndexedObjectType::getPrev(int index, bool mark)
{
    bool wrapped = false;
    while (1) {
        if (--index <= 0) {
            index = max;
            if (wrapped)
                return 0;
            wrapped = 1;
        }
        if (isExistant(index) && (!mark || getObject(index).isMarked()))
            return index;
    }
}

int
GIndexedObjectType::getCount() const
{
    // FIXME: do we need this method? We could simplify (=remove) init if we don't.
    return count;
}
