/**
  *  \file int/names.cc
  *  \brief Variable Name Mapping
  *
  *  (c) 2001-2009 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/names.h"
#include "io/streams.h"
#include "io/io.h"
#include "util/except.h"

/**************************** IntVariableNames ***************************/

const IntVariableNames::index_t IntVariableNames::nil;

/** Constructor. Build blank mapping. */
IntVariableNames::IntVariableNames()
{ }

IntVariableNames::~IntVariableNames()
{ }

/** Look up name.
    \param name Variable name
    \return Index, nil if not found */
IntVariableNames::index_t
IntVariableNames::lookup(string_t name) const
{
    for (std::vector<string_t>::size_type i = 0; i < names.size(); ++i)
        if (name == names[i])
            return i;
    return nil;
}

/** Add a mapping. Note that this does not attempt to avoid duplicates.
    \param name Variable name
    \return Assigned index */
IntVariableNames::index_t
IntVariableNames::add(string_t name)
{
    names.push_back(name);
    return names.size() - 1;
}

/** Add a mapping unless it is already there. */
IntVariableNames::index_t
IntVariableNames::addMaybe(string_t name)
{
    index_t i = lookup(name);
    if (i == nil)
        i = add(name);
    return i;
}

/** Look up address.
    \param index Index (address), must exist
    \return Associated name
    \pre index < getNumNames() */
string_t
IntVariableNames::lookupAddress(index_t index) const
{
    return names[index];
}

/** Get number of names. */
IntVariableNames::index_t
IntVariableNames::getNumNames() const
{
    return names.size();
}

/** Store name list to stream. */
void
IntVariableNames::save(Stream& s, uint32 count)
{
    ASSERT(count <= names.size());
    for (uint32 i = 0; i < count; ++i)
        storePascalString(s, names[i], false);
}

/** Load name list from stream. */
void
IntVariableNames::load(Stream& s, uint32 count)
{
    while (count--) {
        string_t name;
        if (!getPascalString(s, name, false))
            throw FileTooShortException(s);
        add(name);
    }
}
