/**
  *  \file ui/widget-button.cc
  *  \brief Push-button-like Widgets
  *
  *  (c) 2001-2009 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "ui/widgets.h"
#include "ui/event.h"
#include "gfx/gfxcompl.h"
#include "gfx/gfx.h"
#include "gfx/font.h"
#include "ui/uiglobal.h"
#include "ui/util.h"
#include "gfx/palette.h"
#include "ui/commands.h"
#include "cpluslib/assert.h"

/**************************** UIAbstractButton ***************************/

/** \class UIAbstractButton
    \brief Standard Button

    This class provides the event handling mechanisms of a standard
    clickable button (checkbox, radiobutton, ...). The button can be
    have an associated hot-key if so desired.

    Buttons are triggered by a keystroke, mouse click and possibly a
    command.

    Derived classes must override draw().

    You can attach own handlers to sig_fired. */

/** Construct new key button. The button will post a UIKey event.
    \param key key that triggers this button. */
UIAbstractButton::UIAbstractButton(UIKey key)
    : UIWidget(),
      key(key)
{
    setState(bf_Key, true);
}

/** Construct new button.
    \param id widget id, used for emitting commands etc.
    \param key key that triggers this button */
UIAbstractButton::UIAbstractButton(int id, UIKey key)
    : UIWidget(id),
      key(key)
{ }

UIAbstractButton::~UIAbstractButton()
{ }

UIAbstractButton&
UIAbstractButton::addCommand()
{
    sig_fired.add(this, &UIAbstractButton::postCommand);
    return *this;
}

UIAbstractButton&
UIAbstractButton::addStop()
{
    sig_fired.add(this, &UIAbstractButton::callStop);
    return *this;
}

UIAbstractButton&
UIAbstractButton::addKey()
{
    setState(bf_Key, true);
    return *this;
}

/** Change button state. When the button loses logical focus, it must
    be un-pressed. Certain events also need redraw. */
void
UIAbstractButton::onStateChange(int astate, bool enable)
{
    UIWidget::onStateChange(astate, enable);

    // FIXME: can we get rid of this status manipulation?
    if (astate & st_Disabled)
        status &= ~(bf_Pressed | st_Selected);

    if ((status & (bf_Pressed | st_Selected)) == bf_Pressed)
        status &= ~bf_Pressed;

    if (astate & (bf_Pressed | st_Selected | bf_Highlight | st_Disabled | bf_PressOverride))
        drawWidget(false);
}

/** Event handler. Processes mouse and keyboard events, and calls
    fire() when needed. */
bool
UIAbstractButton::handleEvent(const UIEvent& e, bool second_pass)
{
    switch(e.type) {
     case ev_Key:
        if (!hasState(st_Disabled) && !hasState(bf_Key) && e.key.code == key) {
            setSelectedWidget();
            fire();
            return true;
        }
        break;
     case ev_Mouse:
        if (isInside(GfxPoint(e.mouse.x, e.mouse.y)) && !hasState(st_Disabled)) {
            /* mouse is in widget. */
            setSelectedWidget();
            if (e.mouse.b) {
                /* mouse pressed in button */
                setState(bf_Pressed, true);
            } else if (hasState(bf_Pressed)) {
                /* mouse released in button */
                setState(bf_Pressed, false);
                if (hasState(bf_Key)) {
                    putKey(key | getCurrentModifierKeys());
                }
                fire();
            }
            return true;
        } else {
            /* automatically lose selection when mouse is out. */
            if (hasState(st_Selected))
                clearSelectedWidget();

            /* let others handle the event, too. */
            return false;
        }
        break;
     default:
        break;
    }
    return UIWidget::handleEvent(e, second_pass);
}

bool
UIAbstractButton::handleCommand(const UICommand& cmd, UIBaseWidget& sender)
{
    if (cmd.code == cm_INT_QueryReferencePoint) {
        // FIXME: do we need this? This is the default setting.
        ASSERT(cmd.infop);
        TReferencePointArg& ra = *static_cast<TReferencePointArg*>(cmd.infop);
        ra.x      = getExtent().x;
        ra.y      = getExtent().getBottomY();
        return true;
    } else if (cmd.code == cm_INT_Push) {
        setState(bf_PressOverride, true);
    } else if (cmd.code == cm_INT_Unpush) {
        setState(bf_PressOverride, false);
    }
    return UIWidget::handleCommand(cmd, sender);
}

void
UIAbstractButton::postCommand(int prefix)
{
    handleCommand(UICommand(getId(), prefix), *this);
}

void
UIAbstractButton::callStop(int)
{
    stop(id);
}

/** Reaction on button click. */
void
UIAbstractButton::fire()
{
    sig_fired.raise(consumePrefixArg());
}

/******************************** UIButton *******************************/

/** \class UIButton
    \brief Push button

    This class provides the look of a standard push-button. */

UIButton::~UIButton()
{ }

namespace {
    /* Definition of a button shape. A number consists of a number of components
       that have different count and color depending on the button's state.

                 aaaaaaaaaaaaaaaaaaaaaaaaaaaa    a,b  left/top lines
                 abbbbbbbbbbbbbbbbbbbbbbbbbbc    c    bottom/right line
                 ab.........................c    ...  body (with text)
                 ab.........................c
                 ab.........................c
                 ab.........................c
                 accccccccccccccccccccccccccc

       The shapes defined here are a 99% approximation of the "standard" shape
       used in PCC 1.x. 1.x had the fun config setting of a "button thickness",
       but I omit that here. The setting was a remainder of the original PCC 1.x
       button style which had an even thicker frame. */

    struct ButtonShape {
        uint8 num_left;         // Number of left/top lines
        uint8 left[4];          // Colors of left/top lines
        uint8 num_right;        // Number of right/bottom lines
        uint8 right[4];         // Colors of right/bottom lines
        uint8 body_color;       // Color of button body
    };
}

static const ButtonShape button_shapes[] = {
    // --- Normal Size ---
    // normal
    { 2, { COLOR_WHITE, COLOR_GRAYSCALE+10 }, 2, { COLOR_BLACK, COLOR_GRAYSCALE+8 }, COLOR_GRAYSCALE+9 },
    { 2, { COLOR_WHITE, COLOR_GRAYSCALE+11 }, 2, { COLOR_BLACK, COLOR_GRAYSCALE+9 }, COLOR_GRAYSCALE+10 },
    // normal highlight
    { 3, { COLOR_BLACK, COLOR_GRAYSCALE+6, COLOR_GRAYSCALE+9 },  2, { COLOR_WHITE, COLOR_GRAYSCALE+13 }, COLOR_GRAYSCALE+12 },
    { 3, { COLOR_BLACK, COLOR_GRAYSCALE+7, COLOR_GRAYSCALE+10 }, 2, { COLOR_WHITE, COLOR_GRAYSCALE+14 }, COLOR_GRAYSCALE+13 },
    // pressed
    { 3, { COLOR_BLACK, COLOR_GRAYSCALE+13, COLOR_GRAYSCALE+11 }, 2, { COLOR_GRAYSCALE+3, COLOR_GRAYSCALE+9 },  COLOR_GRAYSCALE+9 },
    { 3, { COLOR_BLACK, COLOR_GRAYSCALE+14, COLOR_GRAYSCALE+12 }, 2, { COLOR_GRAYSCALE+4, COLOR_GRAYSCALE+10 }, COLOR_GRAYSCALE+10 },
    // pressed highlight
    { 3, { COLOR_GRAYSCALE+0, COLOR_GRAYSCALE+5, COLOR_GRAYSCALE+7 }, 1, { COLOR_GRAYSCALE+15 }, COLOR_GRAYSCALE+12 },
    { 3, { COLOR_GRAYSCALE+1, COLOR_GRAYSCALE+6, COLOR_GRAYSCALE+8 }, 1, { COLOR_GRAYSCALE+15 }, COLOR_GRAYSCALE+13 },

    // --- Small Size ---
    // normal
    { 1, { COLOR_WHITE }, 1, { COLOR_BLACK }, COLOR_GRAYSCALE+9 },
    { 1, { COLOR_WHITE }, 1, { COLOR_BLACK }, COLOR_GRAYSCALE+10 },
    // normal highlight
    { 1, { COLOR_WHITE }, 1, { COLOR_BLACK }, COLOR_GRAYSCALE+12 },
    { 1, { COLOR_WHITE }, 1, { COLOR_BLACK }, COLOR_GRAYSCALE+13 },

    // pressed
    { 2, { COLOR_BLACK, COLOR_GRAYSCALE+9 },  1, { COLOR_WHITE }, COLOR_GRAYSCALE+9 },
    { 2, { COLOR_BLACK, COLOR_GRAYSCALE+9 },  1, { COLOR_WHITE }, COLOR_GRAYSCALE+10 },
    // pressed highlight
    { 2, { COLOR_BLACK, COLOR_GRAYSCALE+12 }, 1, { COLOR_WHITE }, COLOR_GRAYSCALE+12 },
    { 2, { COLOR_BLACK, COLOR_GRAYSCALE+12 }, 1, { COLOR_WHITE }, COLOR_GRAYSCALE+13 },
};

/** Draw a button. Exported to be callable from non-button widgets that contain
    button-lookalikes, such as scrollbars.
    \param can        Canvas to draw on
    \param extent     Dimensions of the button
    \param name       Text on button
    \param flags      Style flags (bf_XXX, st_XXX) */
void
UIButton::drawButton(GfxCanvas& can, GfxRect extent, const string_t& name, int flags)
{
    // convert flags into button_shapes index
    int slot = 0;
    if ((flags & bf_SmallSize) != 0)
        slot += 8;
    if ((flags & (bf_Pressed | bf_PressOverride)) != 0)
        slot += 4;
    if ((flags & bf_Highlight) != 0)
        slot += 2;
    if ((flags & st_Selected) != 0)
        slot += 1;

    const ButtonShape* sh = &button_shapes[slot];

    // draw button
    GfxRect e = extent;
    int delta = 0;
    for (int i = 0; i < 4; ++i) {
        if (i < sh->num_left) {
            // upper-left corner
            can.drawHLine(e.x, e.y, e.getRightX(),    standard_colors[sh->left[i]], ~0U, GFX_ALPHA_OPAQUE);
            can.drawVLine(e.x, e.y+1, e.getBottomY(), standard_colors[sh->left[i]], ~0U, GFX_ALPHA_OPAQUE);
            ++e.x;
            ++e.y;
            --e.w;
            --e.h;
            ++delta;
        }
        if (i < sh->num_right) {
            // lower-right corner
            can.drawHLine(e.x, e.getBottomY()-1, e.getRightX(),    standard_colors[sh->right[i]], ~0U, GFX_ALPHA_OPAQUE);
            can.drawVLine(e.getRightX()-1, e.y,  e.getBottomY()-1, standard_colors[sh->right[i]], ~0U, GFX_ALPHA_OPAQUE);
            --e.w;
            --e.h;
            --delta;
        }
    }
    can.drawBar(e, standard_colors[sh->body_color], GFX_TRANSPARENT, SOLID_FILL, GFX_ALPHA_OPAQUE);

    // draw text
    const int fnr = (flags & bf_SmallSize) ? FONT_SMALL : FONT_TITLE;
    GfxParameters parm;
    if (flags & st_Disabled)
        parm.color = standard_colors[COLOR_BLACK];
    else
        parm.color = standard_colors[COLOR_SHIELD + 5];

    int x = extent.x + extent.w/2 - fonts[fnr]->getTextWidth(name)/2 + delta;
    int y = extent.y + extent.h/2 - fonts[fnr]->getTextHeight(name)/2 + delta;
    fonts[fnr]->outText(can, parm, x, y, name);
}

/** Draw the button. */
void
UIButton::drawContent(GfxCanvas& can)
{
    drawButton(can, extent, name, getState());
    can.update(extent);
}

void
UIButton::getLayoutInfo(LayoutInfo& info)
{
    if (hasState(bf_SmallSize)) {
        info.flags = 0;
        info.min_size = info.pref_size
            = GfxPoint(name.length() == 1 ? 15 : 8 + fonts[FONT_SMALL]->getTextWidth(name), 15);
    } else {
        info.flags = 0;
        info.min_size = info.pref_size
            = GfxPoint(name.length() == 1 ? 25 : 15 + fonts[FONT_TITLE]->getTextWidth(name), 25);
    }
}
